import myhdl
import myhdl_lib
import pihdf
from pihdf import MFDesign, StrBuilder, info, warn, err
from pihdf.printers.dotty import print_dotty_file
from pihdf.printers.json import print_json_file

import os
import os.path

import shutil

import webbrowser


class CommandHandler():
    '''|
    | TODO
    |________'''    
    def __init__(self, cl_arguments):

        self.module_path   = cl_arguments["MPATH"]
        self.file_name     = cl_arguments["file_name"]
        self.command       = cl_arguments["<command>"]
        self.arguments     = cl_arguments["<args>"]
        self.opt_Name      = cl_arguments["-n"]
        self.opt_Direction = cl_arguments["-d"]
        self.opt_Push      = cl_arguments["-p"]
        self.opt_Data      = cl_arguments["-w"]
        self.opt_BusType   = cl_arguments["-i"]
        self.opt_ParValue  = cl_arguments["-v"]

        self.full_name = self.module_path + '/' + self.file_name
        
        self.status = 0 # Used to return the status of executing the tests

        # Register command handlers
        self.handle = {
            'new'     : self.on_new,
            'add'     : self.on_add,
            'remove'  : self.on_remove,
            'update'  : self.on_update,
            'test'    : self.on_test,
            'convert' : self.on_convert,
            'clean'   : self.on_clean,
            'delete'  : self.on_delete,
            'webdoc'  : self.on_webdoc,
            'version' : self.on_version
        }

        self.handle[ self.command ]()


    def on_new(self):
        '''|
        | Command new creates a new module project from a .json file located in MPATH.
        | The json file must have the same name as the directory in which it is located.
        | If directory and/or .json file do not exist, this function creates the directory
        | and a default .json file (containing only Reset and Clock interfaces).
        |________'''
        self.check_path_and_create_json( self.full_name )

        gdf = MFDesign()
        gdf.initialize( self.full_name )
        gdf.generate(gdf.module_name + ' (top)')

        if gdf.modules != []:
            sub_modules = gdf.init_submodules()

            print_dotty_file(gdf)

            if sub_modules != []:
                sub_path = gdf.c_path + "/src/modules/"
                os.mkdir(sub_path)
                StrBuilder().write(sub_path + '__init__.py', overwrite=True)
                for m in sub_modules:
                    m.c_path = sub_path + m.module_name
                    m.generate(m.module_name + ' (sub-module)')
                    print_json_file(m)        


    def on_update(self):
        '''|
        | Command update re-generates the files, which are created with command new.
        | This command is useful when, e.g., HW module's interfaces or structure changes.
        | First, the corresponding .json file is modified capturing the changes.
        |________'''
        recursive = "recursive" in self.arguments
        
        if recursive == True:
            files = [os.path.join(root, file) for root, dirs, files in os.walk(self.module_path) for file in files if file.endswith('.json')]
            for file_name in files:
                gdf = MFDesign()
                gdf.update( file_name )
        else:
            if os.path.exists( self.full_name ):
                gdf = MFDesign()
                gdf.update( self.full_name )
            else:
                err("File not found: " + self.file_name)
                info("Did you mean: 'update recursive'?")

                
    def on_add(self):
        '''|
        | Command 'add' adds an interface to the .json file and updates the HW module design files
        |________'''
        interface_type = self.arguments[0]
        
        gdf = MFDesign()
        gdf.initialize( self.full_name )

        found = self.check_name( gdf, self.opt_Name )

        if found:
            err("Can not add - interface with the same name exists: " + self.opt_Name)
            exit(0)

        if interface_type == 'HSD':
            gdf.interfaces.append({'push': self.opt_Push, 'direction': self.opt_Direction, 'data': self.opt_Data, 'type': interface_type, 'name': self.opt_Name})

        elif interface_type == 'Bus':
            gdf.interfaces.append({'interfaces': self.opt_BusType, 'type': interface_type, 'name': self.opt_Name})

        elif interface_type == 'Parameter':
            gdf.parameters.append({'value': self.opt_ParValue, 'name': self.opt_Name})

        elif interface_type == 'STAvln':
            gdf.interfaces.append({'direction': self.opt_Direction, 'width': self.opt_Data, 'type': interface_type, 'name': self.opt_Name})

        else: # this should not happen
            err("Unknown type: " + interface_type)
            exit(0)
            
        gdf.overwrite = True
        print_json_file( gdf )

        gdf.update( self.full_name )
        

    def on_remove(self):
        '''|
        | Command 'remove' removes an interface from the .json file and updates the HW module design files
        |________'''
        interface_name = self.arguments[0]

        gdf = MFDesign()
        gdf.initialize( self.full_name )

        found = self.check_name( gdf, interface_name )

        if not found:
            err("Name not found: " + interface_name)
            exit(0)

        # Remove Interface
        gdf.interfaces[:] = [value for value in gdf.interfaces if value["name"] != interface_name]
        # Remove Parameter
        gdf.parameters[:] = [value for value in gdf.parameters if value["name"] != interface_name]

        gdf.overwrite = True
        print_json_file( gdf )

        gdf.update( self.full_name )
    
        
    def on_test(self):
        '''|
        | Command 'test' executes the unit tests for the given module.
        | The tests are specified in the corresponding utest*.py file. 
        |________'''
        self.remove_all('.pyc')

        file_name = "utest_" + self.file_name[:-5] + '.py'
        
        if "recursive" in self.arguments:
            cmd = 'nosetests -vsx -w ' + self.module_path + ' -i "test"'

        elif "with-coverage" in self.arguments:
            cmd = 'nosetests -vsx ' + self.module_path + '/' + 'test/' + file_name + \
                  " --with-coverage --cover-erase --cover-html" + " --cover-html-dir=" + self.module_path + "/out/coverage_html"

        elif self.arguments != []: # to run specific test
            cmd = 'nosetests -vsx ' + self.module_path + '/' + 'test/' + file_name + ':Test_' + self.file_name[:-4] + self.arguments[0]

        else:
            cmd = 'nosetests -vsx ' + self.module_path + '/' + 'test/' + file_name
            
        status = os.system(cmd)
        self.status = status % 255 # On Unix: the first byte indicates signals, the second byte - the status
        
        
    def on_convert(self):
        '''|
        | Command 'convert' converts the myhdl RTL description of a module to synthesizable verilog or vhdl.
        |________'''
        self.remove_all('.pyc')

        # TODO: convert verilog and convert vhdl as separate commands
        
        cmd = 'python ' + self.module_path + '/' + self.file_name[:-5] + '.py'
        os.system(cmd)


    def on_clean(self):
        '''|
        | Command 'clean' removes the files, which have been generated as a result of running simulation or conversion to verilog or vhdl.
        |________'''
        OK = True
        recursive = "recursive" in self.arguments

        if not os.path.exists( self.full_name ) and not recursive:
            warn("File not found: " + self.file_name)
            choise = raw_input("Proceed with 'clean recursive'? y/N ")
            OK = (choise == 'Y' or choise == 'y')

        if OK:
            self.remove_all(*['.pyc', '.vcd', '.vhd', '.tvr', '.*~'])
            
            for root, dirs, files in os.walk(self.module_path):
                for file in files:
                    if file.endswith('.v') and 'out' in root: # To avoid deleting .v files in the 'src' directory (vhd?)
                        os.remove(root + '/' + file)

                for dir in dirs:
                    if dir == "coverage_html":
                        shutil.rmtree(root + '/' + dir)


    def on_delete(self):
        '''|
        | Command 'delete' removes all files and directories from the module's directory tree. It keeps only the .json file.
        |________'''
        choise = raw_input("Delete design? y/N ")
        OK = (choise == 'Y' or choise == 'y')

        if OK:
            cmd = 'cp ' + self.module_path + '/' + self.file_name + ' ~/' + self.file_name
            os.system(cmd)

            cmd = 'rm -rf ' + self.module_path + '/*'
            os.system(cmd)

            cmd = 'mv ~/' + self.file_name + " " + self.module_path + '/' + self.file_name
            os.system(cmd)


    def on_webdoc(self):
        webbrowser.open_new_tab("http://hnikolov.github.io/pihdf_doc")


    def on_version(self):
        info("myhdl     : " + myhdl.__version__)
        info("myhdl_lib : " + myhdl_lib.__version__)
        info("pihdf     : " + pihdf.__version__)
        info("module    : " + "0.1.1")


    # =======================================================================
    def remove_all(self, *args):
        '''|
        | 'args' is a list of file extensions. This function removes all files with the
        | given extension in all sub-directories of 'module_path'
        |________'''
        for root, dirs, files in os.walk(self.module_path):
            for file in files:
                for extension in args:
                    if file.endswith(extension):
                        os.remove(root + '/' + file)


    def check_path_and_create_json(self, file_name):
        '''|
        | Checks if the new module directory and the corresponding .json file exists in the provided path.
        | If not, this function creates the directory and a default .json file (containing only Reset and Clock interfaces).
        | In this way, a new module design (directory structure and files) can always be initialized correctly.
        |________'''
        dirname, filename = os.path.split(file_name)

        if not os.path.exists( dirname ):
            info("Target module path does not exist. Creating a directory...")
            os.makedirs( dirname )

        if not os.listdir( dirname ): # empty directory
            info("Target module .json file does not exist. Creating a default file...")

        if len(os.listdir( dirname )) > 1:
            err("The target module directory is not empty. Use command 'delete' before you start again. Aborting...")
            exit(0)

        # Create a default .json file
        if not os.path.isfile( file_name ):
            s = StrBuilder()
            s.indent()
            s += '{\n'
            s += s.indent() + '"design":\n'
            s += '{\n'
            s += s.indent() + '"name": "' + filename[:-5] + '",\n'
            s += '"interfaces":\n'
            s += '[\n'
            s.indent()
            s += '{ "name": "rst", "type": "Reset", "active": "1" },\n'
            s += '{ "name": "clk", "type": "Clock" }\n'
            s += s.dedent() + ']\n'
            s += s.dedent() + '}\n'
            s += s.dedent() + '}\n'

            s.write( file_name )

        
    def check_name(self, gdf, name ):
        '''|
        | Used by commands 'add' and 'remove'
        |________'''        
        for value in gdf.interfaces:
            if value["name"] == name:
                return True
        for value in gdf.parameters:
            if value["name"] == name:
                return True
        return False



from myhdl import *

class Schedulable():
    '''|
    | This class executes schedules of driving stimuli and capture outputs during testing
    |________'''
    def __init__(self):
        self.cntr_map = {}

    def count(self, cnt, done, end, N):
        '''|
        | Counts events (data items sent or received). Signals END when a predefined number N is reached
        |________'''
        @instance
        def count_inst():
            end.next = False
            while True:
                if cnt.next == N:
                    end.next = True
                yield done
                cnt.next += 1

        return count_inst


    def evaluate(self, tpls):
        '''|
        | Evaluates a single condition or a list of conditions (ANDed). Yields when the condition evaluates as True.
        | Condition is a tuple: (port_name, data_item_index)
        |     port_name - string, name of a port
        |     data_item_index - int, index of a data item send/received on the port given by port_name
        | A condition evaluates True when the data item port_name[data_item_index] has been sent/received
        |________'''
        if not isinstance(tpls, list):
            tpls = [tpls]

        for name, index in tpls:
            cnt = self.cntr_map[name]
            while not cnt >= index + 1:
                yield cnt


    def en_interface(self, clk, start, ipg, cnt, en, cond_lst):
        '''|
        | Generates drive/capture enable signals based on evaluation of the conditions and inter-packet gap
        |________'''
        @instance
        def enbl():
            yield start

            # default schedule with IPG in case of empty condition list
            if len(cond_lst) == 0:
                i = 0
                while True:
                    while not (cnt >= i):
                        yield cnt

                    # Inter-Packet Gap
                    for _ in range(ipg):
                        yield clk.posedge

                    en.next = not en
                    i += 1
            #--------------------------------------------------
            i = 0
            condition = cond_lst.pop(0)
            while True:
                while not (cnt == i):
                    yield cnt

                ipg_loc = ipg

                if condition[0] == i:
                    yield self.evaluate(condition[1])

                    # Get optional local ipg
                    if len(condition) > 2:
                        ipg_loc = condition[2]

                    if len(cond_lst):
                        condition = cond_lst.pop(0)

                # Inter-Packet Gap
                for _ in range(ipg_loc):
                    yield clk.posedge

                en.next = not en
                i += 1

        return enbl


    def gen(self, rst, clk, shed_param, tst_data, sim_time):
        '''|
        | Generates all the instances needed to implement a scheduler
        | ls_x - list of port parameters, one parameter per port.
        | A parameter is a tuple (port_name, done_signal, enable_signal, ipg)
        |________'''
        ls_counters = []
        ls_checkers = []

        ls_sim_end = []
        cond_sim_end = tst_data["cond_sim_end"]

        start = Signal(bool(0))
        sim_end = Signal(bool(0))

        for name, (done, en, ipg) in shed_param.iteritems():
            cntr = Signal(int(0))
            end = Signal(bool(0))
            ls_sim_end.append(end)

            self.cntr_map[name] = cntr
            ls_counters.append(self.count(cntr, done, end, cond_sim_end[name]))
            ls_checkers.append(self.en_interface(clk, start, ipg, cntr, en, tst_data["cond_" + name]))

        sim_ctrl = self.simulation_control(rst, clk, start, sim_time, sim_end)

        @always_comb
        def sim_done():
            x = True
            for s in ls_sim_end:
                x = x and s
            sim_end.next = x

        return instances()


    def simulation_control(self, rst, clk, en_ctrl, sim_time, sim_end):
        '''|
        | Used to control the simulation:
        |     1) enable the drivers and the scheduler after reset
        |     2) stop the simulation after predefined time or after the scheduler indicates
        |________'''

        @instance
        def simControl():
            yield clk.posedge # rst needs 1 clk to get its value

            if rst.active:
                while rst:
                    yield clk.posedge
            else:
                while not rst:
                    yield clk.posedge

            en_ctrl.next = True

            if sim_time == "auto":
                yield sim_end

                for i in range(10):
                    yield clk.posedge
            else:
                for i in range(sim_time):
                    yield clk.posedge

            # End of simulation
            raise StopSimulation

        return instances()


from str_builder import StrBuilder


def print_utest_file(mfdo):
    '''|
    | Create utest_<module_name>.py file
    |________'''
    s = StrBuilder()
    filename =  mfdo.c_path + '/' + mfdo.test_path + '/utest_' + mfdo.module_name + ".py"
    base_test = 't_' + mfdo.module_name
    s += 'import unittest\n\n'
    s += 'from myhdl_lib import *\n\n'

    s += 'from ' + base_test + ' import ' + base_test + '\n\n'

    s += 'class Test_' + mfdo.module_name + '(' + base_test + '):\n'
    s += s.indent() + s.header_comment('The main class for unit-testing. Add your tests here.')

    s += 'def __init__(self):\n'
    s.indent()

    s += '# call base class constructor\n'
    s += base_test + '.__init__(self)\n\n'

    s += s.dedent() + '# Automatically executed BEFORE every TestCase\n'
    s += 'def setUp(self):\n'
    s += s.indent() + base_test + '.setUp(self)\n\n'

    s += s.dedent() + '# Automatically executed AFTER every TestCase\n'
    s += 'def tearDown(self):\n'
    s += s.indent() + base_test + '.tearDown(self)\n\n\n'

    s += s.dedent() + '# ----------------------------------------------------------------------------\n'
    s += '# @unittest.skip("")\n'
    s += 'def test_000(self):\n'
    s += s.indent() + '""" >>>>>> TEST_000: TO DO: describe the test """\n\n'

    if mfdo.parameters != []:
        s += 'self.dut_params = {'
        for p in mfdo.parameters:
            s += (s.noIndent() + '"' + p["name"] + '":' + p["value"] + ', ')
        s = s-2 + (s.noIndent() + '}\n')

    s += 'self.models = {"top":self.BEH}\n'
    s += '# Set fdump to True in order to generate test vector files for the global interfaces\n'
    s += 'self.tb_config = {"simulation_time":"auto", "cosimulation":False, "trace":False, "fdump":False, "ipgi":0, "ipgo":0}\n\n'

    s += '# TO DO: generate stimuli and reference data here\n\n'
    s + 'self.run_it()\n\n'

    # Commented due to problems in case of structured design using BUS interfaces
    #if mfdo.modules != []:
        #s += s.dedent() + '# ----------------------------------------------------------------------------\n'
        #s += '@unittest.skip("")\n'
        #s += 'def test_0000(self):\n'
        #s += s.indent() + '""" >>>>>> TEST_0000: Test template for using stimuli files """\n\n'

        #if mfdo.parameters != []:
            #s += 'self.dut_params = {'
            #for p in mfdo.parameters:
                #s += (s.noIndent() + '"' + p["name"] + '":' + p["value"] + ', ')
            #s = s-2 + (s.noIndent() + '}\n')

        #s += 'self.models = {"top":self.RTL'
        #for m in mfdo.modules:
            #s += (s.noIndent() + ', "' + m["name"] + '":self.BEH')
        #s += (s.noIndent() + '}\n')

        #s += '# Set fdump to True in order to generate test vector files for the local interfaces\n'
        #s += 'self.tb_config = {"simulation_time":"auto", "cosimulation":False, "trace":False, "fdump":False, "ipgi":0, "ipgo":0}\n\n'

        #s += '# TO DO: Provide stimuli and reference files here\n'

        #for i in mfdo.interfaces:
            #if i["direction"]=='IN':
                #s += 'self.stim_' + i["name"] + '.append({"file" : self.test_path + "/vectors/stim_' + i["name"] + '.tvr"})\n'
            #elif i["direction"]=='OUT':
                #s += 'self.res_' + i["name"] + '.append({"file" : self.test_path + "/vectors/res_' + i["name"] + '.tvr"})\n'
                #s += 'self.ref_' + i["name"] + '.append({"file" : self.test_path + "/vectors/ref_' + i["name"] + '.tvr"})\n'
        #s.newLine()

        #s + 'self.run_it(checkfiles=True)\n\n\n'

    s += s.dedent() + '# ----------------------------------------------------------------------------\n'

    s.write(filename, overwrite = mfdo.overwrite)




# Foreground color sequences:
# GREY   = '\033[90m'
# RED    = '\033[91m'
# GREEN  = '\033[92m'
# YELLOW = '\033[93m'
# BLUE   = '\033[94m'
# PINK   = '\033[95m'
# CIAN   = '\033[96m'
# WHITE  = '\033[97m'
# BLACK  = '\033[98m'

# Background color sequences:
# GREY   = '\033[100m'
# RED    = '\033[101m'
# GREEN  = '\033[102m'
# YELLOW = '\033[103m'
# BLUE   = '\033[104m'
# PINK   = '\033[105m'
# CIAN   = '\033[106m'
# WHITE  = '\033[107m'

# Text attributes
# BOLD       = '\033[1m'
ENDC       = '\033[0m'
UNDERSCORE = '\033[4m'
BLINK      = '\033[5m'
REVERSE    = '\033[7m'

HEADER  = '\033[93m'
OKBLUE  = '\033[94m'
OKGREEN = '\033[92m'
WARNING = '\033[95m'
ERROR   = '\033[91m'


def info( msg ):
    print '[' + OKGREEN + 'INFO' + ENDC + '] ' + msg

def infob( msg ):
    print '[' + OKBLUE  + 'INFO' + ENDC + '] ' + msg

def warn( msg ):
    print '[' + WARNING + 'WARN' + ENDC + '] ' + msg

def err( msg ):
    print '[' + ERROR   + 'ERRR' + ENDC + '] ' + msg

def head( msg ):
    print len(msg)*'-'
    print OKBLUE + msg + ENDC
    print len(msg)*'-'



"""|
| module: a convenient front-end tool of pihdf
|________

Usage:
    module <command>
    module MPATH <command> [<args> ...] [options]

Main argument:
    MPATH             Specifies a HW module root directory

Commands not using MPATH:
    help              Prints this help message
    webdoc            Opens the web documentation
    version           Prints the version of myhdl, pihdf, and module
  
Other commands:
    new               create a new module project from a <module_name>.json file
    add               add a new top-level interface or parameter
    remove            remove a top-level interface or parameter
    update            re-generate the files of a module
    test              execute the unit tests for the specified module. Optional test name can be provided
    convert           convert the myhdl description of a module to verilog
    clean             remove temporary files generated in the module's directory tree
    delete            delete all files and directories from the module's directory tree; keep the .json file

Arguments can/are used with the following commands:
    update  [recursive]
    convert [verilog]       | [vhdl]
    test    [<test_name>    | [recursive]       [with-coverage]]
    add     (interface_type | parameter_type)   <options>
    remove  (interface_name | parameter_name)

Options:
    The following options are used with command 'add'
   -n <name>           Instance name             [default: None]
   -d <direction>      Can be 'IN' or 'OUT'      [default: None]
   -p <push>           Push-type of interface    [default: False]
   -v <value>          Parameter value           [default: None]
   -i <bus_name>       The set of bus interfaces [default: None]
   -w <width/fields>   Interface fields or width [default: None]
   
More information is available at
    http://hnikolov.github.io/pyhdf_doc    Use <CTRL + mouse click> to follow the link

"""
from docopt import docopt

from CommandChecker import CommandChecker
from CommandHandler import CommandHandler


def main():
    arguments = docopt(__doc__)

    #if arguments["help"] == True and arguments["<command>"] != None: # Consider if needed also: module help [<command>]; check also docopt git help example
    if arguments["<command>"] == 'help':
        print __doc__
    else:
        cc = CommandChecker(  arguments   )
        ch = CommandHandler( cc.arguments )
        print "done"
        exit(ch.status)



from str_builder import StrBuilder
from ip_wrapper_gen import GenWrapperFile

def print_wrap_file(mfdo, v_fname=None):
    '''|
    | Create <module_name>_wrp.py file
    | This file is used to integrate third-party verilog module in a MyHDL design
    | The verilog module must have the same interface signals as the generated in this file signals
    | All verilog implementation files must be specified in file 'compile_list.txt' file in order to be included for co-simulation
    |________'''
    if v_fname != None:
        verilog_file = v_fname
    else:
        verilog_file = mfdo.verilog["path"] + '/' + mfdo.verilog["name"] + '.v'

    gwf = GenWrapperFile()
    gwf.initialize(verilog_file)

    s = StrBuilder()
    filename = mfdo.c_path + '/' + mfdo.src_path + '/' + mfdo.module_name + "_wrp.py"

    s += 'from myhdl import *\n\n'
    s += 'def ' + mfdo.module_name + '_wrp('
    s += s.noIndent() + mfdo.printInterfaces()
    s += s.noIndent() + ', INST_NAME):\n\n'

    s += s.indent() + '""" Interface signals """\n'

    for i in mfdo.interfaces:
        j = mfdo.getInterfaceObj(i)
        if len(j) > 1:
            mfdo.printBusInterfaces(s, j, i["name"])

        iname = i["name"] + '_' if len(j) > 1 else ''
        for k in j:
            for sigName in k.get_sig_names():
                s += iname + k.inst_name + "_" + sigName + ","
                s.noIndent()

            stmp = ".get_src_signals() # produce data\n" if k.direction == 1 else ".get_snk_signals() # consume data\n" # 1=Out
            s = s-1 + (s.noIndent() + " = " + iname + k.inst_name + stmp)
    s.newLine()

    s += '# Need this in order to work...\n'
    s += '@always(' + mfdo.Clock["name"] + '.posedge, ' + mfdo.Reset["name"] + ')\n'
    s += 'def pass_thru():\n'
    s += s.indent() + 'pass\n\n'
    s.dedent()

    gwf.genTheWrapper(s)

    s.write(filename, overwrite = mfdo.overwrite)

            
def print_verilog_file(mfdo):
    '''|
    | Create <module_name>.v file
    | It is an empty (template) verilog file containing the interface signals, which must be used to integrate a
    | verilog module in a MyHDL design. The integration is done by using the auto-generated verilog wrapper file
    | The verilog module implementation has to be provided in this, and possibly other, files by the designers
    | Additional verilog files must be specified in file 'compile_list.txt' in order to be included for co-simulation
    |________'''
    s = StrBuilder()
    filename =  mfdo.c_path + '/' + mfdo.src_path + '/' + mfdo.module_name + '.v'

    s += 'module ' + mfdo.module_name
    if mfdo.parameters != []:
        s.newLine()
        s += s.indent(2) + '#(\n'
        for p in mfdo.parameters:
            s += 'parameter ' + p["name"] + ' = ' + p["value"] + ',\n'
        s = s-2 + '\n'
        s += ')(\n'
    else:
        s += ' (\n'
        s.indent(2)

    if mfdo.Reset != None:
        s += 'input  wire \t\t' + mfdo.Reset["name"] + ',\n'
    else:
        print "Warning: RESET signal not specified!"

    if mfdo.Clock != None:
        s += 'input  wire \t\t' + mfdo.Clock["name"] + ',\n'
    else:
        print "Warning: CLOCK signal not specified!"

    for i in mfdo.interfaces:
        s.newLine()
        j = mfdo.getInterfaceObj(i)
        if len(j) > 1:
            s += '// ' + i["name"] + ' (Bus interface)\n'
            
        # print the interface signals
        iname = i["name"] + '_' if len(j) > 1 else ''
        for k in j:
            itype   = ' (consume data)\n' if k.direction == 0 else ' (produce data)\n'
            s += '// ' + iname + k.inst_name + itype
            
            for sigName, sigLen in zip(k.get_sig_names(), k.get_sig_lens()):
                str_dir = 'input  ' if k.direction == 0 else 'output '
                if sigName=='ready':
                        str_dir = 'input  ' if str_dir=='output ' else 'output '

                sigLenStr = '    \t' if sigLen==1 else '[' + str(sigLen-1) + ':0]\t'
                s += str_dir + 'wire ' + sigLenStr + iname + k.inst_name + '_' + sigName + ',\n'
        
    s = s-2 + '\n'
    s += ');\n\n'
    s.dedent(2)

    s += '/* Custom code begin */\n'
    s += s.noIndent() + mfdo.extractText(filename, "/* Custom code begin */", "/* Custom code end */")
    s += "/* Custom code end */\n\n"
    
    s += 'endmodule\n'
    
    s.write(filename, overwrite = mfdo.overwrite)

    print_wrap_file(mfdo, filename)




from myhdl import *
from str_builder import StrBuilder

import os
import sys


class GenWrapperFile(object):
    '''|
    | This class is used to generate a python wrapper of a verilog design
    |________'''

    def __init__(self):

        self.module_name = ''
        self.interface = []  # keeps the order of the declarations
        self.parameters = []
        self.inputs = []
        self.outputs = []


    def generateWrapperFile(self):
        '''|
        | Generate <module_name_wrapper>.py file
        |________''' 
        print "\nGenerating .py wrapper file."   
  
        s = StrBuilder()
        
        self.genWrapperInterface(s)
        
        s += '# Need this in order to work...\n'
        # We assume that the clock is 'clk'!!!!!
        s += '@always(clk.posedge)\n'
        s += 'def pass_thru():\n'
        s += s.indent() + 'pass\n\n'       
        s.dedent()
 
        self.genTheWrapper(s)
        self.genConvertFunc(s)
        
        filename = self.module_name + '_wrp.py'
        s.write(filename)


    def genWrapperInterface(self, s):
        '''|
        | Generate the interfaces of the wrapper file
        |________'''
        s += 'from myhdl import *\n\n'          
        s += 'def ' + self.module_name + '_wrapper(\n'
 
        s.indent(5)
        for i in self.interface: # to preserve the order
             if i['type'] != 'parameter':
                 s += i['name'] + ',\n'
 
        s+= 'INST_NAME = "' + self.module_name.upper() + '",\n'
        
        for p in self.parameters:
             s += p['name'] + ' = ' + p['value'] + ',\n'           
        s -= 2 
        s += s.noIndent() + '):\n\n'
        
        s.dedent(3)
       

    def genTheWrapper(self, s, py_name=None):
        '''|
        | Generate the wrapper
        |________'''    
        s += '#---------------------------------------------#\n'
        s += '# Define the interface to the verilog ip core #\n'
        s += '#---------------------------------------------#\n'
        str_name = self.module_name if py_name == None else py_name
        s += str_name + '_wrp.verilog_code = \\' + '\n'
        if self.parameters != []:
            s += '"""' + self.module_name + '#(\\n""" + \\' +  '\n'
            
            for p in self.parameters:
                s += '"""    .' + p["name"] + '($' + p["name"] + '),\\n""" + \\' + '\n'
            s = s-11 + (s.noIndent() + '\\n""" + \\' + '\n')
            
            s += '"""    ) $INST_NAME (\\n""" + \\' + '\n'
        else:
            s += '"""' + self.module_name + ' $INST_NAME (\\n""" + \\' + '\n'
                
        for i in self.interface:
            if i["type"] != "parameter":
                s += '"""    .' + i["name"] + '($' + i["name"] + '),\\n""" + \\' + '\n'
        s = s-11 + (s.noIndent() + '\\n""" + \\' + '\n')
        s += '""");"""\n\n'
         
        s += '#-------------------------------------------------------#\n'
        s += '# output, needed when converting the wrapper to verilog #\n'
        s += '#-------------------------------------------------------#\n'                
        for o in self.outputs:      
            s += o["name"] + '.driven = "wire"\n'
        s += '\n'
        s += 'return pass_thru\n\n\n'
       

    def genConvertFunc(self, s):
        '''|
        | Generate function convert()
        |________'''
        s += s.noIndent() + 'def convert():\n\n'
        for i in self.parameters:
            s += i['name'] + ' = ' + i['value'] + '\n'           
        s += '\n'
        # Declare signals
        for i in self.interface:
            if i['type'] != 'parameter':
                x = i["size"]
                stype = 'bool(0)'
                if x.startswith('['): 
                    stype = 'intbv(0)'
                    x = x.replace(":0", "+1:")
                s += i['name'] + '= Signal(' + stype + x + ')\n'
        s += '\n'                
        s += 'toVerilog(' + self.module_name + '_wrapper,\n' 
        s.indent(2)
        for i in self.interface: 
            if i['type'] != 'parameter':
                s += i['name'] + ' = ' + i['name'] + ',\n'
 
        for p in self.parameters:
            s += p['name'] + ' = ' + p['name'] + ',\n'           
        s = s-2 + (s.noIndent() + ' )\n\n\n') 
        
        s.dedent(4)        
        s += 'if __name__ == "__main__":\n'
        s += s.indent() + 'convert()\n'      
                

    def initialize(self, filename):
        '''|
        | Initialize the GenWrapperFile object
        |________'''
        with open(filename) as f:
                     
            for line_number, line in enumerate(f):
                w_list = line.split()
                if w_list != []:
    
                    if w_list[0]=='module':
                        self.module_name = w_list[1]
    
                    elif w_list[0] == 'parameter':
                        name = w_list[1].replace(',','')
                        value = w_list[3].replace(',','')
                        self.parameters.append( {'name':name, 'value':value})
                        self.interface.append(  {'name':name, 'value':value, 'type':w_list[0]})
    
                    elif w_list[0] == 'input' or w_list[0] == 'output':
                        name = w_list[3] if w_list[2].startswith('[') else w_list[2] 
                        size = w_list[2] if  w_list[2].startswith('[') else ''
                        if w_list[0] == 'input':
                            self.inputs.append( {'name':name.replace(',',''), 'size':size})
                        elif w_list[0] == 'output':
                            self.outputs.append({'name':name.replace(',',''), 'size':size})
                        self.interface.append(  {'name':name.replace(',',''), 'size':size, 'type':w_list[0]})
    
                    elif w_list[0] == ');': 
                        break


def main( args ):
    
    gwf = GenWrapperFile()      
    gwf.initialize(args.file_name)  
    gwf.generateWrapperFile()

          
import argparse
if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Create a myhdl wrapper file.')   
    parser.add_argument('-f', '--file', dest='file_name', default="",
                       help='top-level verilog (.v) file')
    args = parser.parse_args()
    
    main( args )          


import myhdl
from myhdl import *

class Reset(myhdl.ResetSignal):
    '''|
    | The reset signal type of pihdf
    | It inherits the myhdl Reset signal and keeps the reset's instance name
    |________'''
    def __init__(self, name=None, active=1, val=1, async=True):
        myhdl.ResetSignal.__init__(self, val, active, async)
        self.inst_name = name
        self.class_name = self.__class__.__name__

    """ Reset pulse (synchronous with the rising edge of the clk """
    def pulse(self, clk, RST_Length_cc=10):
        @instance
        def _reset():
            self.next = self.active
            for i in range(RST_Length_cc):
                yield clk.posedge
            self.next = not self.active
        return _reset


class Clock(myhdl.SignalType):
    '''|
    | The clock signal type of pihdf
    | It inherits myhdl SignalType and keeps the clk's instance name
    |________'''
    def __init__(self, name=None, val=1):
        myhdl.SignalType.__init__(self, val=bool(val))
        self.inst_name = name
        self.class_name = self.__class__.__name__

    """ Clock generator """
    def gen(self, CLK_Period_ns=1): # default period (1 == 20ns)
        @instance
        def _clock():
            self.next = False
            while True:
                yield delay(CLK_Period_ns)
                self.next = not self.val
        return _clock


class Parameter(object):
    '''|
    | The parameter type of pihdf. It has a name and a value
    |________'''
    def __init__(self, name=None, value=None):
        self.inst_name = name
        self.value = value
        self.class_name = self.__class__.__name__

    def __str__(self):
        return self.inst_name



from myhdl import *
from pihdf import mylog
from myhdl_lib import fifo, arbiter

from collections import OrderedDict

import traceback
import os, sys


class HSD(object):
    '''|
    | A generic hand-shake interface used for (safe) communication between HW modules.
    | The interface has source and sink sides. Each side has ready, valid, and data signals to implement a push-back behavior.
    | The source side is used to write data to the interface: when ready is active, drive data and valid.
    | The sink side is used to consume data from the interface: consume data when it is valid, drive ready accordingly.
    | Data width is configurable via a constructor argument.
    | The interface may contain a FIFO buffer connecting the source and sink sides. The FIFO depth is configurable via method 'gen()'
    | The interface supports also a push behavior, i.e., the data from the source side is propagated to the sink side assuming that
    | the consumer of the data is always ready. Useful when implementing control interfaces.
    |________'''
    def __init__(self, data=None, name=None, direction=None, buf_size=0, push=False, terminate=False, filedump=None, lo=-2):

        if name == None:
            (filename,line_number,function_name,text)=traceback.extract_stack()[lo]
            self.inst_name = text[text.find('.')+1:text.find('=')].strip()  # skip 'self.'
            if self.inst_name=='':
                self.inst_name = text[:text.find('=')].strip()
        else:
            self.inst_name = name

        if direction != None:
            self.direction = direction # IN (SNK) or OUT (SRC)

        self.class_name = self.__class__.__name__

        self.buf_size = buf_size
        self.push = push
        self.terminate = terminate
        self.filedump = False
        
        if filedump != None:
            self.filename = filedump + self.inst_name + ".tvr" # to dump test vectors
            # deletes the content of the file if present
            with open(self.filename, "wb") as f:
                self.filedump = True

        ready_default_value = push or terminate

        self.src_ready  = Signal(bool(ready_default_value))
        self.src_valid  = Signal(bool(0))

        if buf_size > 0:
            self.snk_ready  = Signal(bool(ready_default_value))
            self.snk_valid  = Signal(bool(0))
        else:
            self.snk_ready  = self.src_ready
            self.snk_valid  = self.src_valid

        if data != None:
            if isinstance(data, int):
                if data > 1:
                    self.data = [("data", intbv(0)[data:])]
                elif(data == 1):
                    self.data = [("data", bool(0))]
                else:
                    assert "HSD: Non-positive data width: {}".format(data)
            else:
                self.data = data

            self._src_signals = []
            self._snk_signals = []
            self._sig_names   = []
            self._sig_lens    = []

            for f_name, f_type in self.data:
                src_sig = Signal(f_type)
                snk_sig = Signal(f_type)

                self._src_signals.append( src_sig )

                if buf_size > 0:
                    self._snk_signals.append( snk_sig )

                self._sig_names.append(f_name)
                self._sig_lens.append(len(src_sig))

            if buf_size == 0:
                self._snk_signals = self._src_signals

            self.srcDataList = [] # for DRIVING signals
            self.snkDataList = [] # for READING signals

    @staticmethod
    def _assign_signal(aIn, aOut):
        """ Helper function, generates a combinatorial signal assignment """
        @always_comb
        def asgn():
            aOut.next = aIn
        return asgn

    #========================================================================
    def gen(self, rst=None, clk=None):
        '''|
        | Implementation of the interface, FIFO is instantiated between the src and snk data
        |________'''
        assert( not(self.push and self.buf_size > 0) ), \
        "Interface '{:}' ({:}) of type 'push' can not use a fifo buffer! Currently set to size {:}.".\
        format(self.inst_name, self.class_name, self.buf_size)

        #assert( not(not(hasattr(self, 'data')) and self.buf_size > 0) ), \
        #"Interface '{:}' ({:}) with no data can not use a fifo buffer! Currently set to size {:}.".\
        #format(self.inst_name, self.class_name, self.buf_size)

        if self.buf_size > 0 and not self.push:
            #--------------------
            # Handshake with fifo
            #--------------------
            src_ready = self.src_ready
            src_valid = self.src_valid
            src_data = ConcatSignal(*self._src_signals) if len(self._src_signals) > 1 else self._src_signals[0]

            snk_ready = self.snk_ready
            snk_valid = self.snk_valid
            snk_data = Signal(intbv(0)[len(src_data):])

            sig_full, sig_mty = [Signal(bool(0)) for i in range(2)]

            @always_comb
            def assign():
                src_ready.next = not sig_full
                snk_valid.next = not sig_mty

            inst_fifo = fifo( rst   = rst
                            , clk   = clk
                            , we    = src_valid
                            , din   = src_data
                            , full  = sig_full
                            , re    = snk_ready
                            , dout  = snk_data
                            , empty = sig_mty
                            , depth = self.buf_size)

            widths = self._sig_lens
            hi = [sum(widths[i:]) for i in range(len(widths))]
            lo = [hi[i]-widths[i] for i in range(len(widths))]
            slices = [(snk_data(hi[i],lo[i]) if (widths[i] != 1) else snk_data(lo[i])) for i in range(len(widths))]

            x = []
            for i, slc in enumerate(slices):
                x.append(self._assign_signal(slc, self._snk_signals[i]))

        # Always perform the call to instances() - seems to preserve the hierarchy
        return instances()


    #========================================================================
    def get_src_signals(self):
        '''|
        | Returns the source signals
        |________'''
        return [self.src_ready, self.src_valid] + self._src_signals


    def get_snk_signals(self):
        '''|
        | Returns the sink signals
        |________'''
        return [self.snk_ready, self.snk_valid] + self._snk_signals


    def get_sig_names(self):
        '''|
        | Returns the signals' names
        |________'''
        return ['ready', 'valid'] + self._sig_names


    def get_sig_full_names(self):
        '''|
        | Returns the signals' names prepended with the instance name
        |________'''
        return [self.inst_name + "_" + name for name in self.get_sig_names()]


    def get_sig_lens(self):
        '''|
        | Returns the signals' lens
        |________'''
        return [1, 1] + self._sig_lens


    def get_all_signals(self):
        '''|
        | Returns OrderedDictionaly of all interface snk or src signals, depending on the interface direction
        |________'''
        names = self.get_sig_full_names()
        signals = []
        if self.direction==0:  # IN
            signals = (self.get_snk_signals())
        elif self.direction==1:  # OUT
            signals = (self.get_src_signals())
        else:
            raise ValueError, "Unknown direction %d".format(self.direction)

        return OrderedDict(zip(names, signals))

    #========================================================================
    def assign(self, ready, valid, *args):
        '''|
        | Assigns all interface signals on a side snk/src depending on the interface direction
        |________'''
        if self.direction==0:  # IN
            return self.assignRX(ready, valid, *args)
        elif self.direction==1:  # OUT
            return self.assignTX(ready, valid, *args)
        else:
            raise ValueError, "Unknown direction %d".format(self.direction)


    def assignRX(self, rx_ready, rx_valid, *args):
        """|
        | Connect interface signals. Used in function Convert()
        |________"""
        snk_ready = self.snk_ready
        snk_valid = self.snk_valid
        snk_data  = self._snk_signals

        @always_comb
        def assign_rx():
            snk_valid.next = rx_valid
            rx_ready.next  = snk_ready

        x = []
        for i, s in enumerate(args):
            x.append(self._assign_signal(s, snk_data[i]))

        return instances()


    def assignTX(self, tx_ready, tx_valid, *args):
        '''|
        | Connect interface signals. Used in function Convert()
        |________'''
        src_ready = self.src_ready
        src_valid = self.src_valid
        src_data  = self._src_signals

        @always_comb
        def assign_tx():
            src_ready.next = tx_ready
            tx_valid.next  = src_valid

        x = []
        for i, s in enumerate(args):
            x.append(self._assign_signal(src_data[i], s))

        return instances()


    #========================================================================
    def fdump(self, filename=""):
        '''|
        | Set file dump
        |________'''
        self.filename = filename

        # deletes the content of the file if present
        with open(self.filename, "wb") as f:
            self.filedump = True


    def getStimuli(self, inData):
        '''|
        | Get stimuli data from file
        |________'''
        stim_data_rx = []
        pld = {}
        x = self._sig_names
        try:
            with open(inData[0]["file"], 'rb') as myfile:
                for line in myfile.readlines():
                    for name,pckt in zip(x, line.split()):
                        pld[name] = int(pckt)
                    stim_data_rx.append(pld)
                    pld = {}
        except IOError:
            mylog.err("Stimuli file not found for " + self.inst_name + ":" + inData[0]["file"])

        return stim_data_rx

    #========================================================================
    def append(self, val, use_dict=False):
        '''|
        | Add a stimuli packet to the source data list
        |________'''
        if use_dict:
            y = []
            for name in self._sig_names:
                y.append(val[name])
            self.srcDataList.append( (y) )
        else:
            if len(self._sig_names) > 1:
                self.srcDataList.append( (val) )
            else:
                self.srcDataList.append( ([val]) )

    def get(self, use_dict=False):
        '''|
        | Returns a packet from the sink data list
        | Before calling, first check the list with hasPacket()
        | If filedump is set, this function also stores the data to file
        |________'''
        fdata_list = ''
        while len(self.snkDataList):
            pkt_tpl = self.snkDataList.pop(0)

            fdata_list = ' '.join(str(i) for i in pkt_tpl)
            if self.filedump:
                with open(self.filename, "a") as f:
                    f.write(fdata_list + '\n')

            if use_dict:
                y = {}
                for i, name in enumerate(self._sig_names):
                    y[name] = pkt_tpl[i]

                return y
            else:
                return list(pkt_tpl) if len(pkt_tpl) > 1 else pkt_tpl[0]


    def hasPacket(self):
        '''|
        | Returns true if the sink data list contains a packet
        | In order to provide correct behavior, this function
        | can be overridden by child interface classes
        |________'''
        return len(self.snkDataList) > 0


    #========================================================================
    def driver(self, clk, enable, done, tst_data):
        '''|
        | Instantiates a proper interface driver depending on the interface direction
        |________'''
        if self.direction==0:  # IN
            return self.driveInput(clk, enable, done, tst_data["stim_"+self.inst_name])
        elif self.direction==1:  # OUT
            return self.captureOutput(clk, enable, done, tst_data["res_"+self.inst_name])
        else:
            raise ValueError, "Unknown direction %d".format(self.direction)


    def driveInput(self, clk, enable, done, tst_data):
        '''|
        | Drive input. Used in function simulationControl()
        |________'''
        stim_data_rx = self.getStimuli(tst_data) if tst_data and "file" in tst_data[0] else tst_data

        '''|
        | Drives data from source data list to the interface signals. Not for synthesis
        |________'''
        @instance
        def rxLogic():
            yield enable

            while True:
                '''|
                | The link between unit test and low-level interface timing (fromList)
                |________'''
                if len(stim_data_rx) > 0:
                    self.append(stim_data_rx.pop(0), use_dict=True)
                else:
                    return

                yield self.fromList(clk)

                done.next = not done
                yield enable

        return instances()


    def fromList(self, clk):
        '''|
        | Drives data from source data list to the interface signals. Not for synthesis
        |________'''
        while not len(self.srcDataList): # needed in free-running mode
            yield clk.posedge

        while len(self.srcDataList):
            self.src_valid.next = 1

            src_sigs = self._src_signals
            tpl = self.srcDataList.pop(0)

            for i,val in enumerate(tpl):
                src_sigs[i].next = val

            yield clk.posedge

            # wait the data to be consumed
            while not self.src_ready:
                yield clk.posedge

        # Clear the value of the signals
        self.src_valid.next = 0
        for i,val in enumerate(src_sigs):
            src_sigs[i].next = 0


    def fromListInst(self, clk):
        @instance
        def _x():
            while True:
                yield self.fromList(clk)
        return _x


    def captureOutput(self, clk, enable, done, tst_data):
        '''|
        | Capture output. Used in function simulationControl()
        |________'''
        # Needed when fdump=False and dumping intermediate results (my_xxx.tvr), compare data using files
        if tst_data and "file" in tst_data[0]:
            self.fdump(tst_data[0]["file"])

        '''|
        | Capture data from the interface signals to the sink data list. Not for synthesis
        |________'''
        @instance
        def txLogic():
            yield enable

            while True:
                '''|
                | The link between low-level interface timing (toList) and unittest
                |________'''
                if self.hasPacket(): # hasPacket can be overwritten, e.g. STAvln
                    if tst_data and "file" in tst_data[0]:
                        self.get() # and dump it to file
                    else:
                        tst_data.append(self.get(use_dict=True))

                    done.next = not done
                    yield enable

                yield self.toList(clk)

        return instances()


    def toList(self, clk):
        '''|
        | Capture data from the interface signals to the sink data list. Not for synthesis
        |________'''
        self.snk_ready.next = 1

        yield clk.posedge
        while not self.snk_valid:
            yield clk.posedge

        self.snk_ready.next = 0 # needed when schedule is used

        snk_sigs = self._snk_signals
        l = [int(s) for s in snk_sigs]
        tpl = tuple(l)
        self.snkDataList.append( tpl )


    def toListInst(self, clk):
        @instance
        def _y():
            while True:
                yield self.toList(clk)
        return _y


    #========================================================================
    def enable(self, rst, clk, tx_ready, tx_valid, enable):
        '''|
        | Enable unit: Respects the handshake protocol. Data processing can be performed when enable is '1'
        |        tx_ready, tx_valid   - handshake of the output interface
        |        enable               - output, indicating that the input data (e.g., self.data) is valid and result has to be written to interface 'tx'
        |
        |        self.snk_ready, self.snk_valid   - handshake of this interface used as an input
        |________'''
        state = Signal(bool(0))

        @always_comb
        def rdy():
            self.snk_ready.next = 0
            enable.next   = 0

            if state == 0 or tx_ready:
                self.snk_ready.next = 1
                enable.next   = self.snk_valid

        @always_seq(clk.posedge, reset=rst)
        def clk_prcs():
            if self.snk_ready:
                state.next    = self.snk_valid
                tx_valid.next = self.snk_valid

        return instances()

    """
    The following code provides extra handshake functionality to be experimented with.
    We should decide whether the code should stay here.
    """

    @staticmethod
    def hs_join(ls_hsi, hso):
        """ (Many-to-one) Synchronizes multiple input handshake interfaces: output is ready when all inputs are ready
            ls_hsi - list of input (ready, valid) tuples
            hso    - a output (ready, valid) tuple
        """
        N = len(ls_hsi)
        ls_hsi_rdy, ls_hsi_vld = zip(*ls_hsi)
        ls_hsi_rdy, ls_hsi_vld = list(ls_hsi_rdy), list(ls_hsi_vld)
        hso_rdy, hso_vld = hso

        @always_comb
        def join_comb():
            all_vld = bool(1)
            for i in range(N):
                all_vld = all_vld and ls_hsi_vld[i]
            hso_vld.next = all_vld
            for i in range(N):
                ls_hsi_rdy[i].next = all_vld and hso_rdy

        return join_comb


    @staticmethod
    def hs_fork(hsi, ls_hso):
        """ (One-to-many) Broadcasts to multiple output handshake interfaces: input is ready when all outputs are ready
            hsi    - a input (ready, valid) tuple
            ls_hso - list of output (ready, valid) tuples
        """
        N = len(ls_hso)
        hsi_rdy, hsi_vld = hsi
        ls_hso_rdy, ls_hso_vld = zip(*ls_hso)
        ls_hso_rdy, ls_hso_vld = list(ls_hso_rdy), list(ls_hso_vld)

        @always_comb
        def fork_comb():
            all_rdy = bool(1)
            for i in range(N):
                all_rdy = all_rdy and ls_hso_rdy[i]
            hsi_rdy.next = all_rdy
            for i in range(N):
                ls_hso_vld[i].next = all_rdy and hsi_vld

        return fork_comb


    @staticmethod
    def hs_mux(sel, ls_hsi, hso):
        """ (Many-to-one) Multiplexes multiple input handshake interfaces
            ls_hsi - list of input (ready, valid) tuples
            hso    - a output (ready, valid) tuple
            sel    - selects an input handshake interface to be connected to the output
        """
        N = len(ls_hsi)
        ls_hsi_rdy, ls_hsi_vld = zip(*ls_hsi)
        ls_hsi_rdy, ls_hsi_vld = list(ls_hsi_rdy), list(ls_hsi_vld)
        hso_rdy, hso_vld = hso

        @always_comb
        def mux_comb():
            hso_vld.next = 0
            for i in range(N):
                ls_hsi_rdy[i].next = 0
                if i == sel:
                    hso_vld.next = ls_hsi_vld[i]
                    ls_hsi_rdy[i].next = hso_rdy

        return mux_comb


    @staticmethod
    def hs_demux(sel, hsi, ls_hso):
        """ (One-to-many) Demultiplexes to multiple output handshake interfaces
            hsi    - a input (ready, valid) tuple
            ls_hso - list of output (ready, valid) tuples
            sel    - selects an output handshake interface to connect to the input
        """
        N = len(ls_hso)
        hsi_rdy, hsi_vld = hsi
        ls_hso_rdy, ls_hso_vld = zip(*ls_hso)
        ls_hso_rdy, ls_hso_vld = list(ls_hso_rdy), list(ls_hso_vld)

        @always_comb
        def demux_comb():
            hsi_rdy.next = 0
            for i in range(N):
                ls_hso_vld[i].next = 0
                if i == sel:
                    hsi_rdy.next = ls_hso_rdy[i]
                    ls_hso_vld[i].next = hsi_vld

        return demux_comb


    @staticmethod
    def hs_arbmux(rst, clk, ls_hsi, hso, sel, ARBITER_TYPE="priority"):
        """ (Many-to-one) Arbitrates multiple input handshake interfaces
            Arbitrated only between inputs that currently have active valid
            ls_hsi - list of input (ready, valid) tuples
            hso    - a output (ready, valid) tuple
            sel    - output that indicates the currently selected input handshake interface
            ARBITER_TYPE - selects the type of arbiter to be used, priority or roundrobin
        """
        ls_vld = [hs[1] for hs in ls_hsi]
        sho_rdy, hso_vld = hso
        s = Signal(intbv(0), min=0, max=len(ls_vld))
        prio_update = Signal(bool(0))

        @always_comb
        def comb():
            prio_update.next = sho_rdy and hso_vld
            sel.next = s

        if (ARBITER_TYPE == "priority"):
            arb = arbiter.priority(ls_vld, s)
        elif (ARBITER_TYPE == "roundrobin"):
            arb = arbiter.roundrobin(rst, clk, ls_vld, s, prio_update)
        else:
            assert "hs_arbmux: Unknown arbiter type: {}".format(ARBITER_TYPE)

        mux = HSD.hs_mux(s, ls_hsi, hso)

        return instances()


    @staticmethod
    def hs_arbdemux(rst, clk, hsi, ls_hso, sel, ARBITER_TYPE="priority"):
        """(One-to-many) Arbitrates to multiple output handshake interfaces
            Arbitrates only between autputs that currently have active ready
            hsi    - a input (ready, valid) tuple
            ls_hso - list of output (ready, valid) tuples
            sel    - output that indicates the currently selected output handshake interface
            ARBITER_TYPE - selects the type of arbiter to be used, priority or roundrobin
        """
        shi_rdy, hsi_vld = hsi
        ls_rdy = [hs[0] for hs in ls_hso]
        s = Signal(intbv(0), min=0, max=len(ls_rdy))
        prio_update = Signal(bool(0))

        @always_comb
        def comb():
            prio_update.next = shi_rdy and hsi_vld
            sel.next = s

        if (ARBITER_TYPE == "priority"):
            arb = arbiter.priority(ls_rdy, s)
        elif (ARBITER_TYPE == "roundrobin"):
            arb = arbiter.roundrobin(rst, clk, ls_rdy, s, prio_update)
        else:
            assert "hs_arbdemux: Unknown arbiter type: {}".format(ARBITER_TYPE)

        demux = HSD.hs_demux(s, hsi, ls_hso)

        return instances()


    @staticmethod
    def data_mux(sel, ls_di, do):
        """ Multiplexes a list of input signals to a output signal """
        N = len(ls_di)
        @always_comb
        def mux_comb():
            do.next = 0
            for i in range(N):
                if i == sel:
                    do.next = ls_di[i]
        return mux_comb

    @staticmethod
    def data_demux(sel, di, ls_do):
        """ Demultiplexes an input signal to a list of output signals """
        N = len(ls_do)
        @always_comb
        def demux_comb():
            for i in range(N):
                ls_do.next = 0
                if i == sel:
                    ls_do.next = di
        return demux_comb

    @staticmethod
    def ls_data_mux(sel, lsls_di, ls_do):
        """ Multiplexes a list of input signal structures to an output structure. A structure is represented by a list of signals """
        N = len(ls_do)
        lsls_in = zip(*lsls_di)
        m = [HSD.data_mux(sel, lsls_in[i], ls_do[i]) for i in range(N)]
        return instances()

    @staticmethod
    def ls_data_demux(sel, ls_di, lsls_do):
        """ Demultiplexes an input signal structure to list of output structures. A structure is represented by a list of signals """
        N = len (ls_di)
        lsls_out = zip(*lsls_do)
        d = [HSD.data_demux(sel, ls_di[i], lsls_out[i])for i in range(N)]
        return instances()


    def join(self, ls_rx):
        """ Joins (synchronizes) a list of HSD instances to the self. The HSD data is not manipulated """
        ls_hs = [hs_if.get_snk_signals()[:2] for hs_if in ls_rx]
        hs_join = HSD.hs_join(ls_hs ,(self.src_ready, self.src_valid))
        return instances()


    def fork(self, ls_tx):
        """ Forks (broadcasts) self to a list of HSD instances. The HSD data is not manipulated """
        ls_hs = [hs_if.get_src_signals()[:2] for hs_if in ls_tx]
        hs_fork = HSD.hs_fork((self.snk_ready, self.snk_valid), ls_hs)
        return instances()


    def mux(self, sel, ls_rx, EXCLUDE_DATA=False):
        """ Multiplexes a list of HSD instances to self """
        ls_hs = [hs_if.get_snk_signals()[:2] for hs_if in ls_rx]
        hs_mux = HSD.hs_mux(sel, ls_hs ,(self.src_ready, self.src_valid))
        if not EXCLUDE_DATA:
            ls_data = [hs_if.get_snk_signals()[2:] for hs_if in ls_rx]
            data_mux = HSD.ls_data_mux(sel, ls_data, self.get_src_signals()[2:])
        return instances()


    def demux(self, sel, ls_tx, EXCLUDE_DATA=False):
        """ Demultiplexes self to a list of HSD instances """
        ls_hs = [hs_if.get_src_signals()[:2] for hs_if in ls_tx]
        hs_demux = HSD.hs_demux(sel, (self.snk_ready, self.snk_valid), ls_hs)
        if not EXCLUDE_DATA:
            ls_data = [hs_if.get_src_signals()[2:] for hs_if in ls_tx]
            data_demux = HSD.ls_data_demux(sel, self.get_snk_signals()[2:], ls_data)
        return instances()


    def arbmux(self, rst, clk, ls_rx, sel, ARBITER_TYPE="priority", EXCLUDE_DATA=False):
        """ Arbitrates a list of HSD instances to self """
        ls_hs = [hs_if.get_snk_signals()[:2] for hs_if in ls_rx]
        hs_arbmux = HSD.hs_arbmux(ls_hs, (self.src_ready, self.src_valid), sel, ARBITER_TYPE)
        if not EXCLUDE_DATA:
            ls_data = [hs_if.get_snk_signals()[2:] for hs_if in ls_rx]
            data_mux = HSD.ls_data_mux(sel, ls_data, self.get_src_signals()[2:])
        return instances()


    def arbdemux(self, rst, clk, ls_tx, sel, ARBITER_TYPE="priority", EXCLUDE_DATA=False):
        """ Arbitrates self to a list of HSD instances """
        ls_hs = [hs_if.get_src_signals()[:2] for hs_if in ls_tx]
        hs_arbdemux = HSD.hs_arbdemux(rst, clk, (self.snk_ready, self.snk_valid), ls_hs, sel, ARBITER_TYPE)
        if not EXCLUDE_DATA:
            ls_data = [hs_if.get_src_signals()[2:] for hs_if in ls_tx]
            data_demux = HSD.ls_data_demux(sel, self.get_snk_signals()[2:], ls_data)
        return instances()




import simplejson
from str_builder import StrBuilder
import pihdf


def print_design(mfdo, s):
    s += '{\n'
    s += s.indent() + '"design":\n'
    s += '{\n'
    s += s.indent() + '"name": "' + mfdo.module_name + '",\n'
    s += '"interfaces":\n'
    s += '[\n'
    s.indent()
    if mfdo.Reset != None:
        s += simplejson.dumps(mfdo.Reset) + ',\n'

    if mfdo.Clock != None:
        s += simplejson.dumps(mfdo.Clock) + ',\n'

    for i in mfdo.interfaces:
        s += simplejson.dumps(i) + ',\n'

    s = s-2 + (s.noIndent() + '\n')
    s += s.dedent() + ']\n'

    if mfdo.parameters != []:
        s = s-1 + (s.noIndent() + ',\n') # insert a comma
        s += '"parameters":\n'
        s += '[\n'
        s.indent()

        for p in mfdo.parameters:
            s += simplejson.dumps(p) + ',\n'

        s = s-2 + (s.noIndent() + '\n')
        s += s.dedent() + ']\n'

    s += s.dedent() + '}\n'
    
def print_structure(mfdo, s):
    s = s-1 + (s.noIndent() + ',\n') # insert a comma
    s += '"structure":\n'
    s += '{\n'
    s.indent()

    if mfdo.local_interfaces != []:
        s += '"local_interfaces":\n'
        s += '[\n'
        s.indent()
        for i in mfdo.local_interfaces:
            s += simplejson.dumps(i) + ',\n'

        s = s-2 + (s.noIndent() + '\n')
        s += s.dedent() + '],\n'

    if mfdo.modules != []:
        s += '"design_modules":\n'
        s += '[\n'
        s.indent()
        for i in mfdo.modules:
            s += '{\n'
            s.indent()
            s += '"name": "' + i["name"] + '", "type": "' + i["type"] + '",\n'
            s += '"path": "' + i["path"] + '",\n'

            s += '"connections":\n'
            s += '[\n'
            s.indent()
                            
            for j in i["connections"]:
                sd = '"direction": "' + j["direction"] + '", ' if "direction" in j else ''
                sn = '"local_name": "' + j["local_name"] + '", '
                sc = '"connect_to": "' + j["connect_to"] + '"'

                s += '{' + sd + sn + sc + '},\n'

            s = s-2 + '\n'
            s += s.dedent() + '],\n'

        s = s-2 + (s.noIndent() + '\n')
        s += s.dedent() + '}\n'

    s += s.dedent() + ']\n'
    s += s.dedent() + '}\n'


def print_json_file(mfdo):
    '''|
    | Create/Update a .json file containing the description of a module (interfaces as well as topology if present)
    |________'''
    s = StrBuilder()
    filename =  mfdo.c_path + '/' + mfdo.module_name + '.json'

    print_design(mfdo, s)

    try:
        with open(filename) as f:
            # The .json file exist, we use the existing 'structure' section (DO NOT UPDATE)
            gdf = pihdf.MFDesign()
            gdf.initialize( filename )
            if gdf.modules != []:
                print_structure(gdf, s)

    except IOError as e:
        if mfdo.modules != []:
            print_structure(mfdo, s)

    s += s.dedent() + '}\n'
    
    s.write(filename, overwrite = mfdo.overwrite)


from str_builder import StrBuilder

def print_beh_file(mfdo):
    '''|
    | Create <module_name>_beh.py file. It is an empty (template) file used to model behavior
    |________'''
    s = StrBuilder()
    filename = mfdo.c_path + '/' + mfdo.src_path + '/' + mfdo.module_name + "_beh.py"

    if mfdo.interfaces or mfdo.parameters:
        s += 'def ' + mfdo.module_name + '_beh('
        for i in mfdo.interfaces:
            s += s.noIndent() + i["name"] + ', '
        for p in mfdo.parameters:
            s += s.noIndent() + p["name"] + ', '
        s = s-2 + (s.noIndent() + '):\n')
    else: # we start without a .json file, so the interface and parameter lists are empty
        s += 'def ' + mfdo.module_name + '_beh():\n'

    s += s.indent() + s.header_comment('Specify the behavior, describe data processing; there is no notion of clock. ' + \
                        'Access the in/out interfaces via get() and append() methods. ' + \
                        'The "' + mfdo.module_name + '_beh" function does not return values.')

    s.newLine()        
    s += 'print "Warning: Behavior model not implemented yet!"\n\n'
    s.write(filename)


def update_beh_file(mfdo):
    '''|
    | Update <module_name>_beh.py file containing custom code used to model behavior
    |________'''
    s = StrBuilder()
    filename = mfdo.c_path + '/' + mfdo.src_path + '/' + mfdo.module_name + "_beh.py"

    with open(filename) as f_old:
        for line in f_old:
            if line.startswith("def " + mfdo.module_name):

                if mfdo.interfaces or mfdo.parameters:
                    s += 'def ' + mfdo.module_name + '_beh('
                    for i in mfdo.interfaces:
                        s += s.noIndent() + i["name"] + ', '
                    for p in mfdo.parameters:
                        s += s.noIndent() + p["name"] + ', '
                    s = s-2 + (s.noIndent() + '):\n')
                else: # the interface and parameter lists are empty
                    s += 'def ' + mfdo.module_name + '_beh():\n'

            else:
                s += line

    s.write(filename, overwrite = mfdo.overwrite)



from str_builder import StrBuilder
import os

from .. import mylog

import sys, imp

def print_custom_interfaces_file(mfdo):
    '''|
    | Create 'interfaces.py' file containing custom fields interfaces defined in the .json file.
    | Comand 'update' modifies this file as well.
    |________'''
    if mfdo.custom_interfaces == []: return

    s = StrBuilder()
    filename = mfdo.c_path + "/interfaces.py"

    ss = ''
    indent = 0

    for interface in mfdo.custom_interfaces:
        for tag in interface.keys():
            if tag == 'name':                    
                sss = interface[tag] + ' = ['
                indent = len(sss) * ' '
                ss += sss

            if tag == 'fields': 
                for fields in interface[tag]:
                    ss += '("' + fields["name"] + '", ' + fields["type"] + '),\n' + indent

        ss = ss[:-len(indent)-2] + ']\n' # removes the last comma
        ss += '#-----------------------------------------------------------\n\n'

    s += "from myhdl import *\n\n" 
    s += ss

    s.write(filename, overwrite=True)

    # make the interface visible in pihdf
    custom_interfaces = imp.load_source('interfaces', filename)

    for iname in dir(custom_interfaces):
        itype = getattr(custom_interfaces, iname)
        if isinstance(itype, list):
            sys._getframe(1).f_globals[iname] = itype





import pihdf
from pihdf import info, warn, err

import os
import os.path


class CommandChecker():
    '''|
    | TODO
    |________'''    
    def __init__(self, cl_arguments):

        self.arguments = cl_arguments

        self.arguments["test_recursive"]     = False
        self.arguments["test_with_coverage"] = False

        if self.arguments["MPATH"]:
            self.module_path            = True
            self.module_path_exist      = os.path.exists(self.arguments["MPATH"])

            self.arguments["file_name"] = os.path.basename(os.path.normpath(self.arguments["MPATH"])) + '.json'
            self.arguments.update({"mpath_exist": self.module_path_exist})
        else:
            self.module_path            = False
            self.module_path_exist      = False

            self.arguments["MPATH"]     = ""
            self.arguments["file_name"] = ""
            self.arguments.update({"mpath_exist": False})
        
        # Register command checkers
        self.check =  {
            'new'     : self.new,
            'add'     : self.add,
            'remove'  : self.remove,
            'update'  : self.update,
            'test'    : self.test,
            'convert' : self.convert,
            'clean'   : self.check_MPATH_exist,
            'delete'  : self.check_MPATH_exist,
            'webdoc'  : self.no_check,
            'version' : self.no_check
        }

        self.validate_command_syntax()
        self.check[ self.arguments["<command>"] ]()


    def check_MPATH_exist(self):
        if not self.module_path_exist:
            err("Module path does not exist")
            exit(0)


    def validate_command_syntax(self):
        if self.arguments["<command>"] not in self.check:
            err("Unknown command: " + self.arguments["<command>"])
            # An additional check
            self.check_MPATH_exist()
            exit(0)

        
    def new(self):
        if not self.module_path:
            err("Missing module path")
            info("Usage: module MPATH new")
            exit(0)


    def update(self):
        self.check_MPATH_exist()

        if self.arguments["<args>"] != [] and "recursive" not in self.arguments["<args>"]: # TODO: recursive or all?
            err("Wrong command option: " + str(self.arguments["<args>"]))
            info("Usage: module MPATH update [recursive]")
            exit(0)
            

    def add(self):
        self.check_MPATH_exist()
        if self.arguments["<args>"] == []:
            err("Missing Interface type")
            exit(0)

        if len(self.arguments["<args>"]) != 1:
            err("More arguments given.")
            self.usage_add()

        if self.arguments["-n"] == "None":
            err("Missing name.")
            self.usage_add()

        interface_type = self.arguments["<args>"][0]

        if interface_type == 'HSD':
            self.check_arguments(self.arguments["-d"], self.arguments["-w"])

        elif interface_type == 'Bus':
            self.check_arguments(self.arguments["-i"])

        elif interface_type == 'Parameter':
            self.check_arguments(self.arguments["-v"])

        elif interface_type == 'STAvln':
            self.check_arguments(self.arguments["-d"], self.arguments["-w"])

        else:
            err("Unknown type: " + interface_type)
            exit(0)

    
    def remove(self):
        self.check_MPATH_exist()
        
        if len(self.arguments["<args>"]) != 1:
            err("Wrong command format")
            info("Usage: module MPATH remove NAME")
            exit(0)


    def test(self):
        self.check_MPATH_exist()

        if "recursive" in self.arguments["<args>"]:
            self.arguments["test_recursive"] = True

        if "with-coverage" in self.arguments["<args>"]:
            self.arguments["test_with_coverage"] = True


    def convert(self):
        self.check_MPATH_exist()

        if len(self.arguments["<args>"]) > 1  or  \
               self.arguments["<args>"] != [] and \
                    (("verilog" not in self.arguments["<args>"]) and ("vhdl" not in self.arguments["<args>"])):
                   
            err("Wrong command option: " + str(self.arguments["<args>"]))
            info("Usage: module MPATH convert [verilog | vhdl]")
            exit(0)

        
    def no_check(self):
        pass


    # =======================================================================
    def check_arguments(self, *arg ):
        if 'None' in arg:
            self.usage_add()
            

    def usage_add(self):
        err("Wrong command format")
        info("Usage: module MPATH add <interface_type> -n <name> -d <direction> -w <width/fields> [-p <push>]")
        info("Usage: module MPATH add Bus              -n <name> -i <bus_type>")
        info("       module MPATH add Parameter        -n <name> -v <value>...")
        exit(0)



from str_builder import StrBuilder
from ..          import mylog


def print_gtkw_file(mfdo, cosim = False):
    '''|
    | GTKW file to visualize vcd traces from simulation
    |________'''
    str_cosim = ''
    dut = mfdo.module_name + "_top" + '.'
    if cosim:
        dut = "tb_" + mfdo.module_name + "_top" + ".dut."
        str_cosim = '_cosim'

    s = StrBuilder()
    stmp = str_cosim + ".gtkw" if cosim else ".gtkw"
    filename = mfdo.c_path + '/' + mfdo.test_path + '/' + mfdo.module_name + stmp

    try:
        with open(filename) as f: mylog.info("File '%s' found. It will NOT be updated!" % filename)

    except IOError as e:
        s += '[dumpfile] "' + mfdo.c_path + '/' + mfdo.module_name + '/out/'  + mfdo.module_name + '_top' + str_cosim + '.vcd"\n'
        s += '[savefile] "' + mfdo.c_path + '/' + mfdo.module_name + '/test/' + mfdo.module_name + '_top' + str_cosim + '.gtkw"\n'

        s += "[timestart] 0\n"
        s += "[pos] -1 -1\n"
        s += "[treeopen] tb_" + mfdo.module_name  + "_top" + ".\n"
        s += "[sst_expanded] 1\n"

        s += dut + mfdo.Reset["name"] + "\n"
        s += dut + mfdo.Clock["name"] + "\n"

        # print the interface signals
        for i in mfdo.interfaces:
            j = mfdo.getInterfaceObj(i)
            iname = i["name"] + '_' if len(j) > 1 else ''
            for k in j:
                xtra = 'rtl.' if not cosim else ''
                s += "-\n"

                cname = '_' + k.inst_name.upper() if len(j) > 1 else ''
                if k.direction == 1: s += "-" + i["name"].upper() + cname + " (OUT)\n"
                else:           s += "-" + i["name"].upper() + cname + " (IN)\n"

                for sigName, sigLen in zip(k.get_sig_names(), k.get_sig_lens()):
                    ss = "[" + str(sigLen-1) + ":0]" if (sigLen > 1) else ''
                    s += dut + xtra + iname + k.inst_name + '_' + sigName + ss + "\n"

        s.write(filename, overwrite=True)



from str_builder import StrBuilder


def print_rtl_file(mfdo):
    '''|
    | Create <module_name>_rtl.py file
    | It is an empty (template) file used to specify (synthesizable) implementation at RTL
    | Interface signals are generated as well
    |________'''
    s = StrBuilder()
    filename = mfdo.c_path + '/' + mfdo.src_path + '/' + mfdo.module_name + "_rtl.py"

    s += 'from myhdl import *\n'
    s += 'import pihdf\n'
    s += 'from pihdf import *\n'
    s += 'from myhdl_lib import *\n\n'

    if mfdo.modules != []:
        #s += 'import os\n\n'

        pp_dict = {}
        tt_dict = {}
        for p in mfdo.modules:
            if p["path"]=='':
                if not p["type"] in tt_dict:
                    tt_dict[p["type"]]=1
                    s += 'from modules.' + p["type"] + '.' + p["type"] + ' import ' + p["type"] + '\n'

            elif not p["path"] in pp_dict:
                if not p["type"] in tt_dict:
                    tt_dict[p["type"]] = 1
                    # TODO: Not clear how to deal with the import paths!!!
                    s += 'from ' + p["type"] + '.' +  p["type"] + ' import ' + p["type"] + '\n'
                    #spath = p["path"][3:].replace("/", ".")
                    #s += 'from ' + spath + ' import ' + p["type"] + '\n'
        s += s.noIndent() + '\n'

    s += '#--- Custom code begin ---#\n'
    s += s.noIndent() + mfdo.extractText(filename, "#--- Custom code begin ---#", "#--- Custom code end   ---#")
    s += "#--- Custom code end   ---#\n\n"

    s += 'def ' + mfdo.module_name + '_rtl('
    s += s.noIndent() + mfdo.printInterfaces()
    if mfdo.modules != []:
        s += s.noIndent() + ', IMPL, FDUMP):\n'
    else:
        s += s.noIndent() + '):\n'

    s += s.indent() + s.header_comment('Top-level MyHDL description. This is converted to RTL velilog...')
    s.newLine()
    s += '""" Interface signals """\n'
    for i in mfdo.interfaces:
        j = mfdo.getInterfaceObj(i)
        if len(j) > 1:
            mfdo.printBusInterfaces(s, j, i["name"])

        for k in j:
            stype  = '.get_src_signals() # produce data\n' if k.direction == 1 else '.get_snk_signals() # consume data\n'
            snames = ', '.join(k.get_sig_full_names())
            s += snames + " = " + k.inst_name + stype

    s += s.noIndent() + '\n'

    #-------------------------------------------------------------------------------------------
    # generate an empty (template) file used to specify RTL (synthesizable) implementation
    #-------------------------------------------------------------------------------------------
    if mfdo.modules == []:
        s += '#--- Custom code begin ---#\n'
        s += s.noIndent() + mfdo.extractText(filename, "    #--- Custom code begin ---#", "    #--- Custom code end   ---#")
        s += "#--- Custom code end   ---#\n\n"
        s += 'return all_instances(' + mfdo.Reset["name"] + ', ' + mfdo.Clock["name"] + ')\n'

    else:
        #---------------------------------------------------------------------------------------
        # generate the gen() body (connected components given in the .json file)
        #---------------------------------------------------------------------------------------

        # declaration of the interfaces
        s += '""" Local interfaces """\n'
        for i in mfdo.local_interfaces:
            s_tmp       =                  i["width"]     if("width"  in i)    else \
                          'data='        + i["data"]      if("data"   in i)    else ''
            s_buf_size  = ', buf_size='  + i["buf_size"]  if("buf_size"  in i) else ''
            s_push      = ', push='      + i["push"]      if("push"   in i)    else ''
            s_terminate = ', terminate=' + i["terminate"] if "terminate" in i  else ''
            s += i["name"] + ' = ' + i["type"] + '(' + s_tmp + s_buf_size + s_push + s_terminate + ', filedump=FDUMP)\n'
        s += s.noIndent() + '\n'

        # declaration of the local parameters
        if mfdo.local_parameters != []:
            s += '# Parameters\n'
            for p in mfdo.local_parameters:
                s += p["name"] + ' = ' + p["value"] + '\n'
            s.newLine()

        s += 'if isinstance(IMPL, dict):\n'
        s.indent()
        cc = 1
        for m in mfdo.modules:
            inst_name = m["name"] if "name" in m else 'inst_' + m["type"] + str(cc)
            s += inst_name + '_impl = IMPL["' + inst_name + '"] if "' + inst_name + '" in IMPL else IMPL["top"]\n'
            cc += 1
        s += s.dedent() + 'else:\n'
        s.indent()
        cc = 1
        for m in mfdo.modules:
            inst_name = m["name"] if "name" in m else 'inst_' + m["type"] + str(cc)
            s += inst_name + '_impl = IMPL\n'
            cc += 1
        s.dedent()
        s += s.noIndent() + '\n'

        # instance of the modules (and port map)
        s += '""" Components """\n'
        cc = 1

        for m in mfdo.modules:
            inst_name = m["name"] if "name" in m else 'inst_' + m["type"] + str(cc)
            s += inst_name + ' = ' + m["type"] + '(' + inst_name + '_impl).gen('
            cc += 1
            for key, value in m.iteritems():
                if key != "name" and key != "path" and key != "type" and key != "connections":
                    s.noIndent()
                    s += key + '=' + value + ', '

            s = s-2 + (s.noIndent() + ')\n') # remove the last comma
        s += s.noIndent() + '\n'

        s += '#--- Custom code begin ---#\n'
        s += s.noIndent() + mfdo.extractText(filename, "    #--- Custom code begin ---#", "    #--- Custom code end   ---#")
        s += "#--- Custom code end   ---#\n"

        s.newLine()
        s += 'return all_instances(' + mfdo.Reset["name"] + ', ' + mfdo.Clock["name"] + ')\n'

    s.write(filename, overwrite = mfdo.overwrite)



from str_builder import StrBuilder
from pihdf import mylog

def print_dotty_file(mfdo):
    '''|
    | Create <module_name>.dot, a dotty graph representing the connections between the modules of a structured design
    |________'''
    s = StrBuilder()
    filename =  mfdo.c_path + '/' + mfdo.module_name + ".dot"

    my_graph = []

    for i in mfdo.interfaces:
        if "direction" in i:
            if i["direction"]=="IN":
                my_graph.append([i["name"], "IN", ''])
            elif i["direction"]=="OUT":
                my_graph.append([i["name"], '', "OUT"])
        else: # TODO: Bus interface
            mylog.warn("Top-level interface '{:}' has no direction! This interface will not be present in the .dot file!".format(i["name"]))
            #my_graph.append([i["name"], '', ''])

    for i in mfdo.local_interfaces:
        my_graph.append([i["name"], '', ''])

    # What about parameters?
    # What about single source multiple destinations?
    for m in mfdo.modules:
        for c in m["connections"]:
            for i in my_graph:
                if c["connect_to"] == i[0]:
                    my_position = 1 if c["direction"]=='OUT' else 2
                    i[my_position] = m["name"]

    # Perform some checks on the graph
    errs = False
    for e in my_graph:
        if e[1]=='':
            mylog.warn("Interface '{:}' has destination ({:}) but no source!".format(e[0],e[2]))
            e[1] = "ERR_SRC"
            errs = True

        if e[2]=='':
            mylog.warn("Interface '{:}' has source ({:}) but no destination!".format(e[0],e[1]))
            e[2] = "ERR_DST"
            errs = True

    # Consolidate all in and out edges between two nodes into single in and single out edge
    my_new_graph = []
    my_dict = {}
    for e in my_graph:
        my_name = e[0]
        my_dst  = e[1]
        my_src  = e[2]
        my_new_name = ''
        for ee in my_graph:
            if ee[1] == my_dst and ee[2] == my_src:
                my_new_name += ee[0] + "\\n"

        if not (my_dst,my_src) in my_dict:
            my_dict[(my_dst,my_src)]=1
            my_new_graph.append([my_new_name, my_dst, my_src])

    s += 'digraph ' + mfdo.module_name + ' {\n'
    s += s.indent() + 'bgcolor=white\n'
#         s += 'splines=compound;\n'
#         s += 'splines=true\n'
    s += 'rankdir=LR;\n'
    if errs:
        s += 'node [shape=ellipse, style=filled, color=orange]; ERR_SRC ERR_DST;\n'
    s += 'node [shape=doublecircle, style=filled, color=lightgray]; IN OUT;\n'
    s += 'node [shape=ellipse]; IN OUT;\n'

    for i in my_new_graph:
        s += i[1] + ' -> ' + i[2] + ' [ label = "' + i[0] + '" ];\n'
    s += '}\n'

#         s += 'a1 -> b3 [dir=both color="red:blue"];\n'
#         s += 'b2 -> a3 [dir=none color="green:red;0.25:blue"];\n'

    s.write(filename, overwrite = mfdo.overwrite)



from myhdl import *
from pihdf import mylog

from HSD import HSD

import math

from struct import pack, unpack

class STAvln(HSD):
    '''|
    | ST Avalon Interface:
    | A streaming interface for packed data (ready, valid, sop, eop, empty, data, error).
    | Signals ready and valid are the ready and valid signals of the base HSD class.
    | A packet is defined with start-of-packet (sop) and end-of-packet (eop) signals.
    | Signal 'empty' defines the number of bytes in the last data word, which has to be discarded.
    | The error signal indicates during eop whether the current packet must be discarded.
    |________'''
    def __init__(self, data_width, name=None, direction=None, buf_size=0, terminate=False, filedump=None):

        self.data_width = data_width
        
        WIDTH_BYTE = data_width/8
        emptywidth = math.log(WIDTH_BYTE, 2)
        emptywidth = math.ceil(emptywidth)
    
        stavln_fields = [ ("sop",   bool(0))
                        , ("eop",   bool(0)) 
                        , ("empty", intbv(0)[emptywidth:]) 
                        , ("data",  intbv(0)[data_width:])
                        , ("err",   bool(0))]

        # invoke base class constructor
        HSD.__init__(self, data=stavln_fields, name=name, direction=direction, buf_size=buf_size, terminate=terminate, filedump=filedump, lo=-3)


    def getStimuli(self, inData):
        '''|
        | Get stimuli data from file. Overrides the method of the base HSD class
        |________'''
        stim_data_rx = []
        try:
            with open(inData[0]["file"], 'rb') as myfile:
                for line in myfile.readlines():
                    stim_data_rx.append({"payload":line.strip().decode("hex")})
        except IOError:
            mylog.err("Stimuli file not found for " + self.inst_name + ":" + inData[0]["file"])

        return stim_data_rx


    def unpacked_word(self, word):
        '''|
        | Helper function. Used in append()
        |________'''
        num_bytes = len(word)
        if num_bytes == 1:
            packstr = 'B'
        elif num_bytes == 2:
            packstr = '!H'
        elif num_bytes == 4:
            packstr = '!L'
        elif num_bytes == 8:
            packstr = '!Q'
        return unpack(packstr, word)[0]
    

    def append(self, packet_dict, use_dict=False):
        '''|
        | Add a stimuli packet to the source data list. Overrides the method of the base HSD class
        |________'''
        packet = packet_dict["payload"] if use_dict else packet_dict
        packetlen = len(packet)
        word_width = self.data_width/8
        lastiter = packetlen/word_width + (1 if (packetlen % word_width) else 0)    
        
        l = []
        for i in range(lastiter):
            
            # Default, no errors
            err = False 
             
            sop = (i == 0)
            eop = (i == lastiter-1)
            mty = 0 if not eop else (word_width - (packetlen % word_width)) % word_width
            data = packet[:word_width] + ('\x00' * mty) # + padding if mty > 0
            data = self.unpacked_word(data)
            packet = packet[word_width:]
            self.srcDataList.append((sop, eop, mty, data, err))
#             print "Tuple:", (sop, eop, mty, data, err)


    def packed_word(self, int_val, mty, leftJustified=True):
        '''|
        | Helper function. Used in get()
        |________'''
        word_width = self.data_width/8
        if word_width == 1:
            packstr = 'B'
        elif word_width == 2:
            packstr = '!H'
        elif word_width == 4:
            packstr = '!L'
        elif word_width == 8:
            packstr = '!Q'
        # remove pad bytes
        word = pack(packstr, int_val)
        if leftJustified:
            return word[:len(word)-mty]
        else:
            return word[mty:len(word)]


    def hasPacket(self):
        '''|
        | Checks for a complete packet present in the data list
        |________'''
        for i in range(len(self.snkDataList)):
            eop = self.snkDataList[i][1]
            if eop: return True
        return False


    def get(self, use_dict=False):
        '''|
        | Returns a packet from the sink data list. Overrides the method of the base HSD class
        | Before calling, first check the list with hasPacket()
        |________'''
        started = False
        packet = ''
        fdata = ''
        while len(self.snkDataList):
            pkt_tpl = self.snkDataList.pop(0)
    
            sop  = pkt_tpl[0]
            eop  = pkt_tpl[1]
            mty  = pkt_tpl[2]
            data = pkt_tpl[3]
            err  = pkt_tpl[4] # TODO: err not used

            if sop: 
                started = True
                       
            if started: 
                if eop:
                    data = self.packed_word(data, mty)
                    packet += data
                    fdata += data.encode('hex')
                    if self.filedump:
                        with open(self.filename, "a") as f:
                            f.write(fdata + '\n')    
                    return {"payload":packet} if use_dict else packet
                else: 
                    data = self.packed_word(data, 0)
                    packet += data
                    fdata += data.encode('hex')
            else:
                mylog.err('Wrong framing: ' + self.inst_name + ' (' + self.class_name + ')')



__version__ = "0.1.1"

from convertible import Convertible, all_instances
from testable import Testable
from schedulable import Schedulable
from DutFactory import DutFactory
from mylog import info, infob, warn, err, head

from pihdf import interfaces
from interfaces import Reset, Clock, Parameter, STAvln, Bus, HSD, net_fields, udp_fields, tcp_fields, ipv_fields, group_fields, port_fields, payload_fields, test_fields, AXI4LM, AXI4LS, wd_fields, SBUS

from pihdf import printers
from printers import MFDesign, StrBuilder

__all__ = ["Convertible",
           "all_instances",
           "Testable",
           "Schedulable",
           "DutFactory",
           "info",
           "infob",
           "warn",
           "err",
           "head",
           "Reset",
           "Clock",
           "Parameter",
           "STAvln",
           "Bus",
           "HSD",
           "net_fields",
           "udp_fields",
           "tcp_fields",
           "ipv_fields",
           "group_fields",
           "port_fields",
           "payload_fields",
           "test_fields",
           "AXI4LM",
           "AXI4LS",
           "wd_fields",
           #"interfaces",
           "MFDesign",
           "StrBuilder",
           "printers"
          ]



from myhdl import *
from pihdf import mylog
from pihdf.interfaces.Bus import Bus

import os, sys
import fileinput

class DutFactory(object):
    '''|
    | Returns a simulation instance of a MyHDL function, intended as a DUT in testbench
    | The user selects a simulator: MyHDL, or co-simulation with external HDL simulator, e.g. icarus
    | The user selects whether traces are generated or not
    |________'''

    def __init__(self, cosimulation=False, trace=False):
        '''  '''
        self._simulator = 'icarus' if cosimulation else "myhdl"
        self._trace     =  trace
        self.top_suffix = "_top"

        # Registry of external HDL simulators
        self.sim_reg = {}

        # Register icarus as default external HDL simulator
        self.registerSimulator(
            name         = "icarus",
            hdl          = "Verilog",
            analyze_cmd  = 'iverilog -o {topname}.o -f {modulepath}/src/compile_list.txt',
            simulate_cmd = "vvp -m /.pihdf/myhdl.vpi {topname}.o"
        )

    def registerSimulator(self, name=None, hdl=None, analyze_cmd=None, elaborate_cmd=None, simulate_cmd=None):
        '''|
        | Registers an HDL _simulator
        |        name          - str, user defined name, used to identify this _simulator record
        |        hdl           - str, case insensitive, (verilog, vhdl), the HDL to which the simulated MyHDL code will be converted
        |        analyze_cmd   - str, system command that will be run to analyze the generated HDL
        |        elaborate_cmd - str, optional, system command that will be run after the analyze phase
        |        simulate_cmd  - str, system command that will be run to simulate the analyzed and elaborated design
        |
        |        Before execution of a command string the following substitutions take place:
        |            {topname} is substituted with the name of the simulated MyHDL function
        |________'''
        if not isinstance(name, str) or (name.strip() == ""):
            raise ValueError("Invalid _simulator name")

        if hdl.lower() not in ("vhdl", "verilog"):
            raise ValueError("Invalid hdl {}".format(hdl))

        if not isinstance(analyze_cmd, str) or (analyze_cmd.strip() == ""):
            raise ValueError("Invalid analyzer command")

        if elaborate_cmd is not None:
            if not isinstance(elaborate_cmd, str) or (elaborate_cmd.strip() == ""):
                raise ValueError("Invalid elaborate_cmd command")

        if not isinstance(simulate_cmd, str) or (simulate_cmd.strip() == ""):
            raise ValueError("Invalid _simulator command")

        self.sim_reg[name] = (hdl.lower(), analyze_cmd, elaborate_cmd, simulate_cmd)


    def selectSimulator(self, simulatorName):
        if not simulatorName:
            raise ValueError("No simulator specified")

        if not simulatorName=='myhdl' and not self.sim_reg.has_key(simulatorName):
            raise ValueError("Simulator {} is not registered".format(simulatorName))

        self._simulator = simulatorName


    def enableTrace(self):
        '''|
        | Enables traces (generate a waveform .vcd file)
        |________'''
        self._trace = True


    def disableTrace(self):
        '''|
        | Disables traces
        |________'''
        self._trace = False


    def _getCosimulation(self, module, **kwargs_struct):
        '''|
        | Returns a co-simulation instance of module.
        | Uses the _simulator specified by self._simulator.
        | Enables traces if self._trace is True
        |     module        - MyHDL function to be simulated
        |     kwargs_struct - dict of module structural interface assignments: for signals, interfaces and parameters
        |________'''
        vals = {}
        vals['topname'] = module.get_name() + self.top_suffix
        vals['modulepath'] = module.c_path # absolute path
        hdlsim = self._simulator
        if not hdlsim:
            raise ValueError("No _simulator specified")
        if not self.sim_reg.has_key(hdlsim):
            raise ValueError("Simulator {} is not registered".format(hdlsim))

        hdl, analyze_cmd, elaborate_cmd, simulate_cmd = self.sim_reg[hdlsim]

        # Convert to HDL
        if hdl == "verilog":
            # Filter parameters from kwargs_struct and supply them to convert
            params = {name: kwargs_struct[name] for name in module.get_all_parameters().keys() if name in kwargs_struct.keys()}

            module.convert(hdl=hdl, params=params, verbose=False)
            traceFile = module.c_path + '/' + module.out_path + '/' + "{topname}_cosim".format(**vals)
            if self._trace:
                self._enableTracesVerilog("./tb_{topname}.v".format(**vals), traceFile)
                mylog.infob("Co-simulating... trace will be generated ({}.vcd)".format(traceFile))
            else:
                mylog.infob("Co-simulating... trace will NOT be generated".format(traceFile))

        # TODO: Proper handling of VHDL (some day)
        # elif hdl == "vhdl":
        #     toVHDL(module, **kwargs_struct)

        # Analyze HDL
        os.system(analyze_cmd.format(**vals))

        # Elaborate
        if elaborate_cmd:
            os.system(elaborate_cmd.format(**vals))

        # Copy resets and clocks from the structural interface to flat the interface
        kwargs_flat = {}
        kwargs_flat.update({name : kwargs_struct[name] for name in module.get_all_resets().keys()})
        kwargs_flat.update({name : kwargs_struct[name] for name in module.get_all_clocks().keys()})

        # Flatten structural interfaces and add them to the flat interface
        for name in module.get_all_interfaces().iterkeys():
            kwargs_flat.update(kwargs_struct[name].get_all_signals())

        # Cosimulation
        return Cosimulation(simulate_cmd.format(**vals), **kwargs_flat)


    def _enableTracesVerilog(self, verilogFile, traceFile):
        '''|
        | Enables traces in a Verilog file
        |________'''
        fname, _ = os.path.splitext(traceFile)
        inserted = False
        for _, line in enumerate(fileinput.input(verilogFile, inplace = 1)):
            sys.stdout.write(line)
            if line.startswith("end") and not inserted:
                sys.stdout.write('\n\n')
                sys.stdout.write('initial begin\n')
                sys.stdout.write('    $dumpfile("{}.vcd");\n'.format(fname))
                sys.stdout.write('    $dumpvars(0, dut);\n')
                sys.stdout.write('end\n\n')
                inserted = True


    def getDut(self, module, **kwargs):
        '''|
        | Returns a simulation instance of module.
        | Uses the simulator specified by self._simulator.
        | Enables traces if self._trace is True
        |     module - MyHDL module (function 'gen') to be simulated
        |     kwargs - dict of module interface assignments: for signals and parameters
        |________'''
        if self._simulator=="myhdl":
            if not self._trace:
                sim_dut = module.gen(**kwargs)
                mylog.infob("Simulating... trace will NOT be generated")
            else:
                traceSignals.name = module.get_name() + self.top_suffix
                sim_dut = traceSignals(module.gen, **kwargs)
                mylog.infob("Simulating... trace will be generated ({}.vcd)".format(traceSignals.name))
        else:
            sim_dut = self._getCosimulation(module, **kwargs)

        return sim_dut


    def __call__(self, module, **kwargs):
        return self.getDut(module, **kwargs)



import traceback
import math

from myhdl import *
from myhdl_lib import pipeline_control
from collections import OrderedDict

class CtrlClass:
    ''' '''
    def __init__(self):
        self._ls = []

    def get_instance_name(self):
        ''' Creates a unique name to be associated with an instantiated register '''
        name = []
        i = -4

        while True:
            (filename,line_number,function_name,text)=traceback.extract_stack()[i]
            i -= 1

            if function_name=="gen":  # Skip the gen functions
                continue

            tmp_str = text.split()

            if tmp_str[0] in ['h', '_top', 'sim_dut']: # Extra levels to filter in case of cosimulation
                break

            name.append(tmp_str[0])

        name.reverse()

        return '.'.join(name)


    def get_interface(self, tp, width, sname):
        ''' Creates and returns interface signals. Stores the interface in self._ls '''

        # This name represents the instantiation hierarchy
        inst_name = self.get_instance_name()
        iname = sname if inst_name == "" else sname + '@' + inst_name

        re = Signal(bool(0))
        rd = Signal(intbv(0)[width:])
        we = Signal(bool(0))
        wd = Signal(intbv(0)[width:])

        if tp == "wg":
            for i in self._ls:
                if sname == i["iname"]:
                    we = i["we"]
                    wd = i["wd"]
                    break
            else:
                self._ls.append({"iname": sname, "type":tp, "width":width, "re":re, "rd":rd, "we":we, "wd":wd})
        else:
            self._ls.append({"iname": iname, "type":tp, "width":width, "re":re, "rd":rd, "we":we, "wd":wd})

        if tp == "ro":
            return re, rd
        elif tp in ["wo", "wg"]:
            return we, wd
        elif tp == "rw":
            return re, rd, we, wd
        else:
            print "Unknown register type: {}".format(tp)


    def print_register_list(self, filename):
        s = "-----------\n"
        s += "Register list:\n"
        s += "-----------\n"
        for i,j in enumerate(self._ls):
            s += "{:3} : [{}:] {}\n".format(i, j['width'], j['iname'])
        s += "-----------\n"

        print s

        if filename != "":
            tmpFile = open(filename, 'w')
            tmpFile.write(s)
            tmpFile.close()


    def create_reg_file(self, rst, clk, wd_if, ra, rd, filename):
        ''' Creates a reg file from all interfaces stored in self._ls'''

        self.print_register_list(filename)

        wr_rdy, wr_vld, wr_addr, wr_data = wd_if.get_snk_signals() # consume data
        rd_rdy, rd_vld, rd_data = rd.get_src_signals() # produce data
        ra_rdy, ra_vld, ra_data = ra.get_snk_signals() # consume data

        ls_we = []
        ls_wd = []
        ls_re = []
        ls_rd = []

        for i in self._ls:
            ls_re.append(i["re"])
            ls_rd.append(i["rd"])
            ls_we.append(i["we"])
            ls_wd.append(i["wd"])

        WBUS_WIDTH = len(wr_data)
        NUM_REGS = len(self._ls)

        START_ADDRESS = 0
        ls_addr =[]
        addr_lo = addr_hi = START_ADDRESS

        # Map the registers into the address space
        for r in self._ls:
            num_words = int(math.ceil(float(r["width"])/WBUS_WIDTH))
            addr_hi = addr_lo + num_words
            ls_addr.append(range(addr_lo, addr_hi))
            addr_lo = addr_hi

#         print ls_addr

        print "-----------"
        print "Address map:"
        print "-----------"
        for i,al in enumerate(ls_addr):
            for j,a in enumerate(al):
                lo = j*WBUS_WIDTH
                hi = min((j+1)*WBUS_WIDTH, self._ls[i]["width"])
                print "{:3} : [{:3}:{:3}] {}".format(a, hi, lo, self._ls[i]["iname"])
        print "-----------"

        # Deser registers
        NUM_ADDR = sum([len(ls_addr_per_reg) for ls_addr_per_reg in ls_addr])
        MAX_REG_WRAP =  max([len(ls_addr_per_reg) for ls_addr_per_reg in ls_addr])
#         print "MAX_REG_WRAP =", MAX_REG_WRAP
#         ls_wreg = [Signal(intbv(0)[WBUS_WIDTH:]) for i in range(MAX_REG_WRAP)]
#         wreg = ConcatSignal(*reversed(ls_wreg)) if (len(ls_wreg) > 1) else ls_wreg[0]

        # Create deserialization index map (in which deser register to write: addr -> wreg_sel)
        ls_wreg_idx = tuple([idx for ls_addr_per_reg in ls_addr for idx in range(len(ls_addr_per_reg))])
#         print "ls_wreg_idx  =", ls_wreg_idx

        # Write enable map: at which addresses we have to generate we to the actual register (the last address of a register)
        ls_wreg_we = tuple([1 if idx==len(ls_addr_per_reg)-1 else 0 for ls_addr_per_reg in ls_addr for idx in range(len(ls_addr_per_reg))])
#         print "ls_wreg_we  =", ls_wreg_we

        # Read enable map: at which addresses we have to generate re to the actual register (the first address of a register)
        ls_wreg_re = tuple([1 if idx==0 else 0 for ls_addr_per_reg in ls_addr for idx in range(len(ls_addr_per_reg))])
#         print "ls_wreg_re  =", ls_wreg_re

        # Select map: to which register an address belongs
        ls_wreg_sel = tuple([i for i, ls_addr_per_reg in enumerate(ls_addr) for _ in ls_addr_per_reg])
#         print "ls_wreg_sel  =", ls_wreg_sel


        # Register address High
#         ls_reg_addr_hi = tuple([ls_addr_per_reg[-1] for ls_addr_per_reg in ls_addr])
#         print "ls_reg_addr_hi =", ls_reg_addr_hi

        # Register address Low
#         ls_reg_addr_lo = tuple([ls_addr_per_reg[0] for ls_addr_per_reg in ls_addr])
#         print "ls_reg_addr_lo =", ls_reg_addr_lo

        # Register length in number of address cells occupied
#         ls_reg_len = tuple([len(ls_addr_per_reg)-1 for ls_addr_per_reg in ls_addr])
#         print "ls_reg_len =", ls_reg_len


        def bus_deserializer(rst, clk, wr_rdy, wr_vld, wr_addr, wr_data, ls_we, ls_wd):
            ''' Deserializes bus write access to registers: registers to address space, guarantees atomic register write
                    wr_rdy, wr_vld, wr_addr, wr_data - handshake interface for the write bus
                    ls_we - list of write enable signals, one signal per register
                    ls_wd - list of write data signals, one signal per register
                    Current behavior:
                        - Suppose we have:
                            - 20 bit register X
                            - 8 bit write data bus
                            - X is mapped to bus addresses:
                                X[ 7: 0] -> addr i
                                X[15: 8] -> addr i+1
                                X[19:16] -> addr i+2
                        - To write to X we have to do the following:
                            write(addr=i,   data=d0)
                            write(addr=i+1, data=d1)
                            write(addr=i+2, data=d2)
                            note: write(addr, data) represents bus transactions via the write bus interface (wr_rdy, wr_vld, wr_addr, wr_data)
                        - What happens under the hood:
                            write(addr=i,   data=X[ 7: 0]) - d0=X[ 7: 0] is written in an intermediate temporary register 0
                            write(addr=i+1, data=X[15: 8]) - d1=X[15: 8] is written in an intermediate temporary register 1
                            write(addr=i+2, data=X[19:16]) - d2=X[19:16] is written in an intermediate temporary register 2, and
                                                       d0, d1, d2 are transferred from the temporary registers to X (atomic write)
                            Write to X is triggered by write to addr i+2, the top address of a register. If we just write to i+2, without
                            before that writing to i and i+1, X will still be written with incorrect data from the temporary registers
                            TODO: May be, protect X such that write to it is executed only after i, i+1, i+2 are written sequentially
            '''

            NUM_STAGES  = 1

            stage0_en               = Signal(bool(0))
            stop0_rx                = Signal(bool(0))
            stop0_tx                = Signal(bool(0))

            pipe0_tx_vld = Signal(bool(0))
            pipe0_tx_rdy = Signal(bool(0))

            pipe0_ctrl_i = pipeline_control( rst             = rst,
                                             clk             = clk,
                                             rx_vld          = wr_vld,
                                             rx_rdy          = wr_rdy,
                                             tx_vld          = pipe0_tx_vld,
                                             tx_rdy          = pipe0_tx_rdy,
                                             stage_enable    = stage0_en,
                                             stop_rx         = [stop0_rx],
                                             stop_tx         = [stop0_tx] )

            # Temporary reg where the data is deserialized before writing it to the real register
            ls_treg = [Signal(intbv(0)[WBUS_WIDTH:]) for i in range(MAX_REG_WRAP)]
            treg = ConcatSignal(*reversed(ls_treg)) if (len(ls_treg) > 1) else ls_treg[0]
            # Selects the register being written
            reg_sel = Signal(intbv(0, min=0, max=NUM_REGS))

            """
            =============================================
            = Pipeline Stage 0
            =============================================
            """
            @always(clk.posedge)
            def S00_proc_wreg():
                ''' Deserialize '''
                if (rst):
                    stop0_rx.next    = 0
                    stop0_tx.next    = 1
                    reg_sel.next    = 0
                else:
                    if (stage0_en):
                        stop0_tx.next    = 1
                        for a in range(NUM_ADDR):
                            if (a == wr_addr):
                                # Deser
                                treg_idx = ls_wreg_idx[a]
                                we = ls_wreg_we[a]
                                sel = ls_wreg_sel[a]
                                ls_treg[treg_idx].next = wr_data
                                if (we==1):
                                    reg_sel.next    = sel
                                    stop0_tx.next   = 0

            def slicer(y,x):
                l = len(y)
                @always_comb
                def slicer_comb():
                    y.next = x[l:]
                return slicer_comb

            slc = [slicer(ls_wd[i], treg) for i in range(NUM_REGS)]

            @always_comb
            def we_demux():
                pipe0_tx_rdy.next = 1
                for i in range(NUM_REGS):
                    if (i == reg_sel):
                        ls_we[i].next = pipe0_tx_vld
                    else:
                        ls_we[i].next = 0

            return instances()

        busdeser = bus_deserializer(rst, clk, wr_rdy, wr_vld, wr_addr, wr_data, ls_we, ls_wd)


        def bus_serializer(rst, clk, ra_rdy, ra_vld, ra_data, rd_rdy, rd_vld, rd_data, ls_re, ls_rd):
            ''' Serializes bus read access to registers: maps registers to address space, guarantees atomic register read
                    ra_rdy, ra_vld, ra_data - input address handshake interface
                    rd_rdy, rd_vld, rd_data - output data handshake interface
                    ls_re - list of read enable signals, one signal per register
                    ls_rd - list of read data signals, on signal per register
                    Current behavior:
                        - Suppose we have:
                            - 20 bit register X
                            - 8 bit read data bus
                            - X is mapped to bus addresses:
                                X[ 7: 0] -> addr i
                                X[15: 8] -> addr i+1
                                X[19:16] -> addr i+2
                        - To read X we have to do the following:
                            d0 = read(addr=i)
                            d1 = read(addr=i+1)
                            d2 = read(addr=i+2)
                            note: read(addr) represents bus transactions via the read bus interface (ra_rdy, ra_vld, ra_data, rd_rdy, rd_vld, rd_data)
                        - What happens under the hood:
                            d0 = read(addr=i) - reads the whole X into an intermediate temporary register (atomic read), and
                                                returns d0 = X[ 7: 0]
                            d1 = read(addr=i+1) - reads d1 = X[15: 8] from the intermediate temporary register
                            d2 = read(addr=i+2) - reads d2 = X[19:16] from the intermediate temporary register

                            Reading X is triggered by a read from addr i, the first address of a register. If we just read from address i+1 or i+2
                            without before that reading from address i, we will be reading some dummy data from the intermediate temporary register.
                            TODO: May be, read X into the temporary register every time there is non-sequential access to some of the register addresses.
                            non-sequential access - registers are not accessed sequentially i, i+1, i+2
            '''

            NUM_STAGES  = 1

            stage1_en               = Signal(bool(0))
            stop1_rx                = Signal(bool(0))
            stop1_tx                = Signal(bool(0))

            pipe1_ctrl_i = pipeline_control( rst             = rst,
                                             clk             = clk,
                                             rx_vld          = ra_vld,
                                             rx_rdy          = ra_rdy,
                                             tx_vld          = rd_vld,
                                             tx_rdy          = rd_rdy,
                                             stage_enable    = stage1_en,
                                             stop_rx         = [stop1_rx],
                                             stop_tx         = [stop1_tx] )

            treg = Signal(intbv(0)[MAX_REG_WRAP*WBUS_WIDTH:])
            ls_treg = [Signal(intbv(0)[WBUS_WIDTH:]) for i in range(MAX_REG_WRAP)]

            def assigner(y,x):
                @always_comb
                def assigner_comb():
                    y.next = x
                return assigner_comb

            slc1 = [assigner(ls_treg[i], treg((i+1)*WBUS_WIDTH,i*WBUS_WIDTH)) for i in range(MAX_REG_WRAP)]


            reg_sel = Signal(intbv(0, min=0, max=NUM_REGS))

            """
            =============================================
            = Pipeline Stage 0
            =============================================
            """
            mux_sel = Signal(intbv(0, min=-1, max=MAX_REG_WRAP))

            def padder(y, x):
                @always_comb
                def padder_comb():
                    y.next = x
                return padder_comb

            ls_rdp = [Signal(intbv(0)[MAX_REG_WRAP*WBUS_WIDTH:]) for i in range(NUM_REGS)]
            pad = [assigner(ls_rdp[i], ls_rd[i]) for i in range(NUM_REGS)]

            @always(clk.posedge)
            def S10_proc_rreg():
                if (rst):
                    stop1_rx.next    = 0
                    stop1_tx.next    = 0
                else:
                    if (stage1_en):
                        for i in range(NUM_REGS):
                            ls_re[i].next = 0
                        if (not stop1_rx):
                            mux_sel.next = -1
                            for a in range(NUM_ADDR):
                                if (a == ra_data):
                                    treg_idx = ls_wreg_idx[a]
                                    mux_sel.next = treg_idx
                                    re = ls_wreg_re[a]
                                    sel = ls_wreg_sel[a]
                                    if (re==1):
                                        ls_re[sel].next = 1
                                        reg_sel.next   = sel
                                        stop1_rx.next   = 1
                                        stop1_tx.next   = 1
                        # TODO: Try using hs_mux to handle two data sources: addresses & regs
                        else:
                            treg.next        = ls_rdp[reg_sel]
                            stop1_rx.next    = 0
                            stop1_tx.next    = 0

            @always_comb
            def dat_mux():
                rd_data.next = 0xFFFFFFFF
                for i in range(MAX_REG_WRAP):
                    if (i == mux_sel):
                        rd_data.next = ls_treg[i]

            return instances()

        busser = bus_serializer(rst, clk, ra_rdy, ra_vld, ra_data, rd_rdy, rd_vld, rd_data, ls_re, ls_rd)

        return instances()


class Bus(object):
    '''|
    | A generic Bus class. This class is a container containing interfaces used to specify a bus.
    | Interfaces can be any of the supported interfaces in pihdf.
    | A Bus is defined by a list of 'interfaces' in file 'fields_interfaces.py', part of pihdf.
    |________'''    
    def __init__(self, bus_type=None, name=None, reg_file=False, filedump=None, lo=-2):

        self.interface_list = bus_type()
        self.reg_file = reg_file
                    
        if name == None:
            # Extract the instance name
            (filename,line_number,function_name,text)=traceback.extract_stack()[lo]
            self.inst_name = text[text.find('.')+1:text.find('=')].strip()  # skip 'self.'
            if self.inst_name=='':
                self.inst_name = text[:text.find('=')].strip()
        else:
            self.inst_name = name

        self.class_name = self.__class__.__name__

        if self.reg_file:
            self.ctrl = CtrlClass()

        for i in self.interface_list:
            i.inst_name = self.inst_name + "_" + i.inst_name


    def interfaces(self):
        '''|
        | Return the list of interfaces. If the list contains only one interfaces, then returns the interface itself.
        |________'''
        if len(self.interface_list) == 1:
            return self.interface_list[0]
        return self.interface_list


    def get_all_signals(self):
        '''|
        | Return OrderedDict of all signals of all bus interface.
        |________'''
        return OrderedDict([(name, signal) for interface in self.interface_list for name, signal in interface.get_all_signals().iteritems()])


    def assign(self, *args):
        '''|
        | Assigns all all signals of all bus interface.
        |________'''
        a = args
        ls_assign = []

        for interface in self.interface_list:
            num_signals = len(interface.get_all_signals())
            signals = a[:num_signals]
            ls_assign.append(interface.assign(*signals))
            a = a[num_signals:]

        return ls_assign


    def create_reg_file(self, rst, clk, filename=""):
        if self.reg_file:
            return self.ctrl.create_reg_file(rst, clk, self.interface_list[0], self.interface_list[1], self.interface_list[2], filename)
        else:
            return []




import myhdl
from myhdl import *
from myhdl._instance import _Instantiator
from myhdl._misc import _isGenSeq

from pihdf.interfaces.HSD  import HSD
from pihdf.interfaces.Bus  import Bus
from pihdf.DutFactory import DutFactory
from pihdf.schedulable import Schedulable

from collections import OrderedDict

import mylog

import os
import sys
import shutil
import fileinput
import traceback
import inspect

import time


class Convertible(object):
    '''|
    | Base class for convertible objects
    |________'''
    def __init__(self):

        self.module_name = self.__class__.__name__
        self.is_top = False
        self.top_suffix = "_top"

        self.resets = []
        self.clocks = []
        self.interfaces = []
        self.parameters = []

        self.IN  = 0
        self.OUT = 1

        self.BEH = 0
        self.RTL = 1
        self.VRG = 2

        # Extract the instance name"
        (filename,line_number,function_name,text)=traceback.extract_stack()[-3]
        self.inst_name = text[:text.find('=')].strip()

        # Used to redirect the generated files in a common directory 
        self.c_path        = os.path.dirname(inspect.getfile(self.__class__))
        self.src_path      = 'src'
        self.out_path      = 'out'
        self.test_path     = 'test'
        self.filedump_path = self.c_path  + '/test/vectors/'

        self.fdump = False


    def init_parameters(self, dut_params={}, verbose=True):
        '''|
        | Assign values to parameters. The values are provided from unit tests.
        | If a value for a parameter is not provided, then the default value is used.
        | Default value is always given when creating a Parameter object)
        |________'''
        for def_param in self.get_parameters():
            if def_param.inst_name in dut_params.keys():
                def_param.value = dut_params[def_param.inst_name]
            elif verbose:
                mylog.warn("No initial value provided for parameter '%s'. The default value (%s = %s) will be used." % (def_param, def_param, def_param.value))

    def get_name(self):
        '''|
        | Returns the name of the HW module
        |________'''
        return self.module_name

    def get_resets(self):
        '''|
        | Returns a list of (unique) resets objects
        |________'''
        return self.resets()

    def get_clocks(self):
        '''|
        | Returns a list of (unique) clock objects
        |________'''
        return self.clocks()

    def get_interfaces(self, FDUMP=False):
        '''|
        | Returns a list of (unique) interface objects
        |________'''
        fdump = self.filedump_path if FDUMP else None
        return self.interfaces(fdump)

    def get_parameters(self):
        '''|
        | Returns the list of parameters objects
        |________'''
        return self.parameters # (not defined with lambda)

    def get_parameter_value(self, aParameterName):
        '''|
        | Returns a parameter value given a parameter name
        |________'''
        for i in self.parameters:
            if i.inst_name == aParameterName:
                return i.value


    def get_all_resets(self):
        '''|
        | Returns OrderedDict of all resets (unique)
        |________'''
        return OrderedDict([(rst.inst_name, rst) for rst in self.get_resets()])

    def get_all_clocks(self):
        '''|
        | Returns OrderedDict of all clocks (unique)
        |________'''
        return OrderedDict([(clk.inst_name, clk) for clk in self.get_clocks()])

    def get_all_interfaces(self, FDUMP=False):
        '''|
        | Returns OrderedDict of all interface (unique)
        |________'''
        return OrderedDict([(interface.inst_name, interface) for interface in self.get_interfaces(FDUMP)])

    def get_all_parameters(self, overwrite_params={}, verbose=False):
        '''|
        | Returns OrderedDict of all parameters (unique). Overwrites the parameters given in overwrite_params
        |________'''
        params = OrderedDict([(parameter.inst_name, parameter.value) for parameter in self.get_parameters()])
        for x in params.keys():
            if x in overwrite_params:
                params[x] = overwrite_params[x]
            elif verbose:
                mylog.warn("No initial value provided for parameter '%s'. The default value (%s = %s) will be used." % (x, x, params[x]))
        return params


    def simulationControl(self, tst_data, sim_time, **dut_kwargs):
        '''|
        | Drive values to signals and collect results from simulation
        |________'''
        rst = dut_kwargs[ self.get_resets()[0].inst_name ] # TODO: Works only for 1 reset
        clk = dut_kwargs[ self.get_clocks()[0].inst_name ] # TODO: Works only for 1 clock

        ls_drivers = []
        shed_param = {}
        for interface_name in self.get_all_interfaces().keys():
            interface = dut_kwargs[interface_name] # TODO: check for presence
            enable, done = [Signal(bool(0)) for _ in range(2)]
            ipg = None

            if isinstance(interface, Bus):
                for ii in interface.interfaces():
                    enable, done = [Signal(bool(0)) for _ in range(2)]
                    ls_drivers.append(ii.driver(clk, enable, done, tst_data))
                    ipg = self.ipgi if ii.direction == self.IN else self.ipgo
                    shed_param[ii.inst_name] = (done, enable, ipg)
            else:
                ls_drivers.append(interface.driver(clk, enable, done, tst_data))
                ipg = self.ipgi if interface.direction == self.IN else self.ipgo
                shed_param[interface_name] = (done, enable, ipg)

        sheduler = Schedulable().gen(rst, clk, shed_param, tst_data, sim_time)

        return instances()


    def Testbench(self, tb_config, tst_data, dut_params={}, verbose=False):
        '''|
        | Set-up and run a simulation of a HW module
        | Uses a DutFactory
        |________'''

        self.is_top = True

        # TODO: Dirty patches
        BUS_PATCH = None
        RST_PATCH = None
        CLK_PATCH = None

        trace        = tb_config["trace"]
        cosimulation = tb_config["cosimulation"]
        sim_time     = tb_config["simulation_time"]
        self.ipgi    = tb_config["ipgi"]
        self.ipgo    = tb_config["ipgo"]

        self.fdump = tb_config["fdump"]
        # Used to create test-vector files from behavior top-level model
        FDUMP = self.fdump and self.IMPL == self.BEH

        if len([self.get_interfaces()]) == 0:
            mylog.infob("Nothing to simulate: module contains no interfaces!")
            return

        # Get clocks and instantiate clock generators
        CLK_Period_ns = 1 # clk period = 20ns
        clocks = self.get_all_clocks()
        clk_gen = [clk.gen(CLK_Period_ns) for clk in clocks.values()]
        CLK_PATCH = clocks.values()[-1] # TODO: Dirty patch, keeps the last from the list

        # Get resets and instantiate reset generators
        resets = self.get_all_resets()
        rst_gen = [rst.pulse(CLK_PATCH, 10) for rst in resets.values()]
        RST_PATCH = resets.values()[-1] # TODO: Dirty patch, keeps the last from the list

        # Get interfaces
        interfaces = self.get_all_interfaces(FDUMP)

        # Overwrite default parameters
        parameters = self.get_all_parameters(overwrite_params=dut_params, verbose=True)

        # Add resets, clock, interfaces, and parameter to DUT argument dictionary
        dut_kwargs = {}
        dut_kwargs.update(resets)
        dut_kwargs.update(clocks)
        dut_kwargs.update(interfaces)
        dut_kwargs.update(parameters)


        ## Create DUT factory
        #dutFact = DutFactory()

        ## Configure DUT factory
        #if cosimulation:
            #dutFact.selectSimulator('icarus')
        #else:
            #dutFact.selectSimulator('myhdl')

        #if trace:
            #dutFact.enableTrace()

        # Generate DUT
        #sim_dut = dutFact(self, **dut_kwargs)

        sim_dut = DutFactory(cosimulation, trace).getDut(self, **dut_kwargs)

        sim_control = self.simulationControl(tst_data=tst_data, sim_time=sim_time, **dut_kwargs)

        return instances()


    def Simulate(self, tb_config, tst_data, dut_params={}, verbose=False):
        _start = time.time()

        tb = self.Testbench(tb_config, tst_data, dut_params, verbose)
#         tb = traceSignals(self.Testbench, tb_config, tst_data, dut_params, verbose)
        Simulation(tb).run()

        if verbose:
            mylog.infob("Simulation took %s seconds" % (time.time() - _start))


    def gen(self, **kwargs):
        '''|
        | Function gen(): the top-level implementation method
        |________'''
        if self.IMPL == self.BEH:
            # remove resets from **kwargs
            r_kwargs = dict(kwargs)
            for rst in self.get_resets():
                del r_kwargs[rst.inst_name]

            # remove clocks from **kwargs
            for clk in self.get_clocks():
                del r_kwargs[clk.inst_name]

            logic = []
            clk = kwargs[self.get_clocks()[0].inst_name]

            for x in self.get_interfaces():
                if isinstance(x, Bus):
                    xx = r_kwargs[x.inst_name]
                    for ii in xx.interfaces():
                        if ii.direction == self.IN:
                            logic.append( ii.toListInst( clk ))
                        elif ii.direction == self.OUT:
                            logic.append(ii.fromListInst( clk ))
                        else:
                            raise ValueError
                else:
                    if x.inst_name in r_kwargs:
                        xx = r_kwargs[x.inst_name]
                        # Do not check xx because in case of structural designs,
                        # it does not have direction set (local interface)
                        if x.direction == self.IN: 
                            logic.append( xx.toListInst( clk ))
                        elif x.direction == self.OUT:
                            logic.append(xx.fromListInst( clk ))
                        else:
                            raise ValueError

            @always(clk.posedge)
            def beh_prcs():
                self.funcdict['beh'](**r_kwargs)

            return instances()

        if self.IMPL == self.RTL:
            if self.structural == True:
                FDUMP = self.filedump_path if self.fdump else None
                kwargs.update({"FDUMP":FDUMP, "IMPL":self.models})

            if self.is_top:
                # Instantiate all register files
                rtl = self.funcdict['rtl'](**kwargs)
                ls_rf = []
                rst = kwargs[ self.get_resets()[0].inst_name ] # TODO: Works only for 1 reset
                clk = kwargs[ self.get_clocks()[0].inst_name ] # TODO: Works only for 1 clock
                for name in self.get_all_interfaces().iterkeys():
                    interface = kwargs[name]
                    if isinstance(interface, Bus) and interface.reg_file == True:
                        ls_rf.append(interface.create_reg_file(rst, clk, self.c_path + '/' + self.out_path + '/' + self.module_name + '_mem_map.txt'))
                return instances()

            return self.funcdict['rtl'](**kwargs)


        if self.IMPL == self.VRG:
            kwargs.update({"INST_NAME":self.inst_name})
            return self.funcdict['vrg'](**kwargs)


    def flat2struct(self, **kwargs):
        '''|
        | Adapts flat interface to structural interface
        |________'''

        assign = []

        interfaces = self.get_all_interfaces()

        # Connect top flat signals to interface signals
        for interface in interfaces.itervalues():
            args = [kwargs.get(full_name, None) for full_name in interface.get_all_signals().iterkeys()]
            assign.append(interface.assign(*args))

        # Connect clocks, resets and parameters directly
        gen_kwargs = {}
        gen_kwargs.update(interfaces)
        gen_kwargs.update({ name : kwargs.get(name, None) for name in self.get_all_clocks().keys()})
        gen_kwargs.update({ name : kwargs.get(name, None) for name in self.get_all_resets().keys()})
        gen_kwargs.update({ name : kwargs.get(name, None) for name in self.get_all_parameters().keys()})

        _top = self.gen(**gen_kwargs) # The name _top is detected in the reg_file when hierarchy is extracted

        return instances()


    def convert(self, hdl='verilog', params={}, verbose=True):
        '''|
        | Converts convertible to HDL
        |________'''
        if self.IMPL == self.BEH:
            mylog.err("Behavior models can not be converted to verilog!")
            exit(0)

        self.is_top = True
        # Add resets, clock and overwritten parameters to top interface
        top_kwargs = {}
        top_kwargs.update(self.get_all_clocks())
        top_kwargs.update(self.get_all_resets())
        top_kwargs.update(self.get_all_parameters(params, verbose))

        # Add flatten interface signals to top interface
        for x in self.get_all_interfaces().itervalues():
            top_kwargs.update(x.get_all_signals())

        if verbose: mylog.infob('Converting to {}...'.format(hdl))
        if hdl.lower()=='verilog':
            toVerilog.name = self.get_name() + self.top_suffix
            toVerilog(self.top, **top_kwargs)
        elif hdl.lower()=='vhdl':
            toVHDL.name = self.get_name() + self.top_suffix
            toVHDL(self.top, **top_kwargs)
        else:
            raise ValueError("Unknown HDL: {}".format(hdl))


    def clean(self):
        '''|
        | 'Clean' (moves) generated .v and .vcd files to 'module'/out
        | To be used also when generating test vectors...
        |________'''
        filename = self.module_name + self.top_suffix
        for file in os.listdir('.'):
        
            if os.path.isfile(file):
                if file == filename + '.o':
                    os.remove(file)
                if (file == filename + '.vcd') or (file == filename + '.v') or (file == filename + '.vhd') or (file == 'tb_' + filename + '.v'):
                    shutil.move(file, self.c_path + '/' + self.out_path + '/' + file)


# Static member of class Convertible
Convertible.hsd_object = []

def _isFifoInterface(obj):
    '''|
    | Checks whether a local interface object needs to instantiate a FIFO
    |________'''
    if isinstance(obj, HSD):
        if obj.buf_size > 0:
            return True
    return False

def all_instances(rst, clk):
    '''|
    | Identical to (substitutes) function 'instances() of myhdl._misc'
    | In addition, we add the instances of the 'local' interfaces, which create FIFOs
    |________'''
    f = inspect.currentframe()
    d = inspect.getouterframes(f)[1][0].f_locals
    l = []
    for v in d.values():
        if _isGenSeq(v):
            l.append(v)
        # In addition, we need the FIFO instances, but add them only once
        if _isFifoInterface(v):
            if v not in Convertible.hsd_object:
                l.append(v.gen(rst, clk))
                Convertible.hsd_object.append(v)
    return l



import myhdl
from str_builder import StrBuilder


def print_module_class_file(mfdo):
    '''|
    | Create <module_name>.py file
    |________'''
    s = StrBuilder()
    filename = mfdo.c_path + '/' + mfdo.module_name + ".py"
    #--------------------------------------------------------------------------------------------
    # imports
    #--------------------------------------------------------------------------------------------
    s += 'from myhdl import *\n'
    s += 'from pihdf import Convertible\n'
    s += 'from pihdf.interfaces import *\n\n'

    if mfdo.modules != []:
        # TODO: Needs rethinking in case of structural designs 
        s += 'import sys\n'
        s += 'import os\n'
        s += 'module_path = os.path.dirname(__file__)\n\n'
        # Make third-party-modules 'visible'
        p_dict = {}
        t_dict = {}
        bflg = 0
        for p in mfdo.modules:
            if p["path"]=='':
                bflg += 1
            elif not p["path"] in p_dict:
                # TODO: Not clear how to deal with the import paths!!!
                s += 'lib_path = os.path.abspath(module_path + "/../' +  p["path"] + '")\n'
                s += 'sys.path.append(lib_path)\n'
                p_dict[p["path"]] = 1
        s.newLine()
        if bflg > 1:
            # Make the modules in directory 'modules' visible for each other
            s += "import site\n"
            s += "site.addsitedir(module_path + '/src/modules')\n\n"

    if mfdo.custom_interfaces != []:
        s += 'from interfaces import *\n\n'

    s += 'from ' + 'src.' + mfdo.module_name + '_beh import *\n'
    if mfdo.verilog != {}:
            s += 'from ' + 'src.' + mfdo.module_name + '_wrp import *\n'
    else:
            s += 'from ' + 'src.' + mfdo.module_name + '_rtl import *\n'
    s.newLine()

    s += 'class ' + mfdo.module_name + '(Convertible):\n'
    s += s.indent() + s.header_comment('The design class')

    s += 'def __init__(self, IMPL={}):\n\n'
    s.indent()

    if mfdo.modules != []:
        s += 'self.structural = True\n'
        s += 'self.models = IMPL\n'
    else:
        s += 'self.structural = False\n'
    s.newLine()

    s += 'if isinstance(IMPL, dict):\n'
    s += s.indent() + 'self.IMPL = IMPL["top"] if "top" in IMPL else IMPL\n'
    s += s.dedent() + 'else:\n'
    s += s.indent() + 'self.IMPL = IMPL\n\n'

    s += s.dedent() + '# call base class constructor\n'
    s += 'Convertible.__init__(self)\n\n'

    # TODO: To be extended for any number of reset signals
    s += 'self.resets = lambda : [\n'
    s += s.indent() + 'Reset(name="' + mfdo.Reset["name"] + '", active=' + mfdo.Reset["active"] + ', val=' + mfdo.Reset["active"] + ', async=True)\n'
    s += s.dedent() + ']\n'

    s.newLine()

    # TODO: To be extended for any number of clock signals
    s += 'self.clocks = lambda : [\n'
    s += s.indent() + 'Clock(name="' + mfdo.Clock["name"] + '")\n'
    s += s.dedent() + ']\n'

    s.newLine()

    s += "self.interfaces = lambda fdump : [\n"
    s.indent()

    for i in mfdo.interfaces:
        s_width     = 'data_width='      + i["width"]        + ', ' if "width"      in i else ''
        s_direction = 'direction=self.'  + i["direction"]    + ', ' if "direction"  in i else ''
        s_intfs     = 'bus_type='        + i["interfaces"]   + ', ' if "interfaces" in i else ''
        s_name      = 'name="'           + i["name"] + '"'   + ', ' if "name"       in i else ''
        s_data      = 'data='            + i["data"]         + ', ' if "data"       in i else ''
        s_push      = 'push='            + i["push"]         + ', ' if "push"       in i else ''
        s_regfile   = 'reg_file='        + i["reg_file"]     + ', ' if "reg_file"   in i else ''
        s += i["type"] + '(' + s_width + s_direction + s_intfs + s_name + s_data + s_push + s_regfile
        s += s.noIndent() + 'filedump=fdump),\n'

    s = s-2 + (s.noIndent() + '\n')
    s += s.dedent() + ']\n'
    
    s.newLine()
    
    if mfdo.parameters != []:
        s += '# no lambda here\n'
        s += 'self.parameters = [\n'
        s.indent()

        for p in mfdo.parameters:
            s += 'Parameter("' + p["name"] + '", value=' + p["value"] + '),\n'
        
        s = s-2 + '\n' # remove the last comma
        s += s.dedent() + ']\n'
    else:
        s += 'self.parameters = []\n'

    s.newLine()

    s += '# register implementations used in Convertible.gen()\n'
    s += 'self.funcdict = {\n'
    s.indent()

    s_beh = mfdo.module_name + "_beh"
    s_rtl = mfdo.module_name + "_rtl" if mfdo.verilog == {} else "None"
    s_vrg = mfdo.module_name + "_wrp" if mfdo.verilog != {} else "None"

    s += "'beh': " + s_beh + ",\n"
    s += "'rtl': " + s_rtl + ",\n"
    s += "'vrg': " + s_vrg + "\n"
    s += s.dedent() + '}\n'

    s.dedent()
    s.newLine(2)

    generateTop(mfdo, s)
    s.newLine(2)

    s += s.dedent(3) + 'if __name__ == "__main__":\n\n'

    s += s.indent() + 'import myhdl\n'
    s += 'import pihdf\n\n'

    s += 'pihdf.info("Using MyHDL version %s" % myhdl.__version__)\n'
    s += 'pihdf.info("Using MyFramework version %s" % pihdf.__version__)\n\n'

    s += 'dn = ' + mfdo.module_name + '(IMPL=1)\n'

    sparams = ''
    if mfdo.parameters != []:
        sparams += ', params={'
        for p in mfdo.parameters:
            sparams += '"' + p["name"] + '":' + p["value"] + ', '
        sparams = sparams[:-2] + '}'# remove the last comma
        
    s += 'dn.convert(hdl="verilog"' + sparams + ')\n'
    s += 'dn.convert(hdl="vhdl"'    + sparams + ')\n'
    s += 'dn.clean()\n'

    s.write(filename, overwrite = mfdo.overwrite)


def generateTop(mfdo, s):
    '''|
    | Generate function top()
    |________'''
    s += "def top(self,\n"
    s.indent(2)
    mfdo.printSignals(s)
    s += s.noIndent() + ':\n'
    
    s += s.dedent() + s.header_comment('Provides flat interface to the top-level implementation. This function is given to MyHDL.toVerilog() and MyHDL.toVHDL()')
    s += "x = locals()\n"
    s += "del x['self']\n"
    s += "return self.flat2struct(**x)\n"



import mylog


class Testable():
    '''|
    | The base class for unit-testing
    |________'''

    def __init__(self):

        self.BEH = 0
        self.RTL = 1
        self.VRG = 2

        self.dut_params = {}
        self.models = {}
        self.tb_config = {}

        self.checkfiles = False
        self.err = 0
        self.verbose = False


    def dict2string(self, conf):
        '''|
        | Converts test configuration (dictionary) into a string
        | Also, updates the test configuration by adding all default items if omitted during test creation
        |________'''
        # Default configuration
        configuration = {"simulation_time":"auto", "cosimulation":False, "trace":False, "fdump":False, "ipgi":0, "ipgo":0}

        # Update the configuration with the actual configuration
        for key,value in conf.iteritems():
            configuration[key] = value

        conf.update(configuration)

        # Convert to string
        s = ''
        for key,value in configuration.iteritems():
            if self.verbose: # Can be set in each test
                s += key + " = " + str(value) + ", "
            else:
                if str(value) != "False":
                    if str(value) == "True":
                        s += key + ", "
                    elif value != 0 and str(value) != "auto":
                        s += key + " = " + str(value) + ", "
                    
        return s[:-2]


    def params2string(self, params):
        '''|
        | Converts parameter information (dictionary) into a string
        |________'''
        s = ''
        for key,value in params.iteritems():
            s += key + " = " + str(value) + ", "

        return s[:-2]        


    def impl2string(self, conf):
        '''|
        | Converts a dictionary (containing implementation information) into a string
        |________'''
        impl     = [self.BEH, self.RTL, self.VRG]
        impl_str = ['BEHAVIOR', 'MyHDL_RTL', '(external) VERILOG']
        s = ''
        for key,value in conf.iteritems():
            if len(conf) == 1:
                return impl_str[value]
            
            if value not in impl:
                mylog.err(key + ": Wrong implementation model selected: (" + str(value) + ")! Exit...")
                exit()
            s += key + " = " + impl_str[value] + ", "
        return s[:-2]


    def check_config(self, cname):
        '''|
        | Check test configuration
        | Determine the end-of-simulation conditions: len of stim_ and ref_ lists
        |________'''
        mylog.info(cname + ": " + self.impl2string(self.models))

        conf_str = self.dict2string(self.tb_config)
        if conf_str != '':
            mylog.info("Configuration: " + conf_str)

        if self.dut_params != {}:
            mylog.info("DUT parameters: " + self.params2string(self.dut_params))

        mty_stim_list = True
        stim_not_present = True
        for key,val in self.tst_data.iteritems():
            if "stim" in key:
                stim_not_present = False
                if val != [] and "file" in val[0]:
                    num_payloads = sum(1 for line in open(val[0]["file"]))
                    self.cond_sim_end[key[5:]] = num_payloads
                else:
                    self.cond_sim_end[key[5:]] = len(val)

                if self.cond_sim_end[key[5:]] > 0:
                    mty_stim_list = False
                
        if stim_not_present or mty_stim_list:
            mylog.info("No stimuli provided")
            
        no_ref_data = True
        for key,val in self.ref_data.iteritems():
            if val[0] != [] and "file" in val[0][0]:
                num_payloads = sum(1 for line in open(val[0][0]["file"]))
                self.cond_sim_end[key] = num_payloads
            else:
                self.cond_sim_end[key] = len(val[0])

            if self.cond_sim_end[key] > 0:
                no_ref_data = False

        # To avoid endless simulation
        if self.tb_config["simulation_time"] == "auto" and (stim_not_present or mty_stim_list) and no_ref_data:
            self.tb_config["simulation_time"] = 5
            mylog.warn("Simulation time ca not be determined! Selecting a constant simulation time.")


    def check_results(self):
        '''|
        | Call different check function depending on the configuration
        |________'''
        if self.checkfiles:
            self.compare_ref_with_res_files()
        else:
            if self.err > 0:
                self.compare_almost_ref_with_res()
            else:
                self.compare_ref_with_res()

        
    def compare_ref_with_res(self):
        '''|
        | Check the results
        |________'''
        for name, (ref,res) in self.ref_data.items():
            assert( len(ref)==len(res)), "\tCOMPARE ERROR: Length mismatch for \"{:}\". Detected lengths: \n\t len(expected)={:} \n\t len(detected)={:} \n\t".format(name, len(ref), len(res))
            # check fields
            for i in range(len(ref)):
                for fld in ref[i].keys():
                    if fld=="payload":
                        exp_pld = ref[i]["payload"].encode("hex") if isinstance(ref[i]["payload"], str) else ref[i]["payload"]
                        det_pld = res[i]["payload"].encode("hex") if isinstance(res[i]["payload"], str) else res[i]["payload"]
                        assert (exp_pld==det_pld), "\tCOMPARE ERROR: \"{:}\" payload #{:}: Payload does not match: \n\t expected={:} \n\t detected={:}".format(name, i+1, exp_pld, det_pld)
                    else:
                        assert (ref[i][fld]==res[i][fld]), "\tCOMPARE ERROR: \"{:}\" fields #{:}: Field \"{:}\" does not match: expected={:} detected={:}".format(name, i, fld, ref[i][fld], res[i][fld])


    #def compare_comp_ref_with_res(self):
        #'''|
        #| TODO: Check the results (complex numbers)
        #|________'''
        #for name, (ref,res) in self.ref_data.items():
            #assert( len(ref)==len(res)), "\tCOMPARE ERROR: Length mismatch for \"{:}\". Detected lengths: \n\t len(expected)={:} \n\t len(detected)={:} \n\t".format(name, len(ref), len(res))
            ## check fields
            #for i in range(len(ref)):
                #for fld in ref[i].keys():
                    #if fld=="payload":
                        #exp_pld = ref[i]["payload"].encode("hex") if isinstance(ref[i]["payload"], str) else ref[i]["payload"]
                        #det_pld = res[i]["payload"].encode("hex") if isinstance(res[i]["payload"], str) else res[i]["payload"]
                        ##self.assertAlmostEqual(exp_pld.real, det_pld.real, 3)
                        ##self.assertAlmostEqual(exp_pld.imag, det_pld.imag, 3)
                        #assert (exp_pld==det_pld), "\tCOMPARE ERROR: \"{:}\" payload #{:}: Payload does not match: \n\t expected={:} \n\t detected={:}".format(name, i+1, exp_pld, det_pld)
                    #else:
                        #assert (ref[i][fld]==res[i][fld]), "\tCOMPARE ERROR: \"{:}\" fields #{:}: Field \"{:}\" does not match: expected={:} detected={:}".format(name, i, fld, ref[i][fld], res[i][fld])


    def compare_almost_ref_with_res(self):
        '''|
        | Check the results (tolerate a difference within an error)
        |________'''
        #err = 0.005 # 0.5%
        for name, (ref,res) in self.ref_data.items():
            assert( len(ref)==len(res)), "\tCOMPARE ERROR: Length mismatch for \"{:}\". Detected lengths: \n\t len(expected)={:} \n\t len(detected)={:} \n\t".format(name, len(ref), len(res))
            # check fields
            for i in range(len(ref)):
                for fld in ref[i].keys():
                    if fld=="payload":
                        exp_pld = ref[i]["payload"].encode("hex") if isinstance(ref[i]["payload"], str) else ref[i]["payload"]
                        det_pld = res[i]["payload"].encode("hex") if isinstance(res[i]["payload"], str) else res[i]["payload"]

                        gt = exp_pld if exp_pld > det_pld else det_pld
                        tm = abs(exp_pld - det_pld)
                        df = tm/float(gt)
                        assert (df <= self.err), "\tCOMPARE ERROR: \"{:}\" payload #{:}: Payload difference too much: \n\t expected={:} \n\t detected={:} \n\t difference={:} (> {:})".format(name, i+1, exp_pld, det_pld, df, self.err)

                    else:
                        exp_pld = ref[i][fld]
                        det_pld = res[i][fld]
                        gt = exp_pld if exp_pld > det_pld else det_pld
                        tm = abs(exp_pld - det_pld)
                        df = tm/float(gt)
                        assert (df <= self.err), "\tCOMPARE ERROR: \"{:}\" fields #{:}: Field \"{:}\" difference too much: expected={:} detected={:} \n\t difference={:} (> {:})".format(name, i+1, exp_pld, det_pld, df, self.err)

                        
    def compare_ref_with_res_files(self):
        '''|
        | Check the results from files
        |________'''
        for name, (ref,res) in self.ref_data.items():
            try:
                with open(ref[0]["file"], "rb") as expfile, open(res[0]["file"], "rb") as resfile:
                    linesref = expfile.readlines()
                    linesres = resfile.readlines()
                    assert(len(linesref)==len(linesres)), "\tCOMPARE ERROR: Length mismatch for \"{:}\". Detected lengths: \n\t len(expected)={:} \n\t len(detected)={:} \n\t".format(name, len(linesref), len(linesres))
                    k=1
                    for lineref, lineres in zip(linesref, linesres):
                        assert (lineref.strip()==lineres.strip()), "\tCOMPARE ERROR: \"{:}\" payload #{:}: Payload does not match: \n\t expected={:} \n\t detected={:}".format(name, k, lineref.strip(), lineres.strip())
            except IOError:
                mylog.err("Compare reference with result files: File not found")
                


from str_builder import StrBuilder
import os

from .. import mylog

def print_compile_list_file(mfdo):
    '''|
    | Create 'compile_list.txt' file containing a default list of verilog files used in case of co-simulation
    | All verilog files, which are part of the MyHDL design but are not auto-generated must be specified in this 'compile_list.txt' file
    |________'''
    s = StrBuilder()
    filename = mfdo.c_path + '/' + mfdo.src_path + '/compile_list.txt'
    filepath = os.getcwd() + '/' + mfdo.c_path + '/' + mfdo.src_path

    s += mfdo.module_name + '_top.v\n'

    if mfdo.verilog != {}:
        ls = extractFileNames(filename, mfdo)

        if ls == []:
            if mfdo.verilog["path"] == '':
                s += filepath + '/' + mfdo.module_name + '.v\n'
            else:
                s += filepath + '/' + mfdo.verilog["name"] + '.v\n'
        else:
            for l in ls:
                s += filepath + l + '\n'

    s += 'tb_' + mfdo.module_name + '_top.v\n'
    s.write(filename, overwrite = mfdo.overwrite)


def extractFileNames(filename, mfdo):
    '''|
    | Extract the names of the 'additional' custom verilog files used to implement the verilog design
    | Assumes that the files are located inside the module's source directory
    |________'''
    ls = []
    try:
        with open(filename) as f:
            for line in f:
                if mfdo.src_path in line:
                    ls.append( line.partition(mfdo.src_path)[2].strip() )

    except Exception as e:
        mylog.err("in extractFileNames():")
        print "      ", e

    return ls



""" setuptools distribution and installation script. """

from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


setup( name                 = "pihdf",
       version              = "0.1.1",
       description          = 'Hardware Design Framework based on python and MyHDL',
       long_description     = readme(),
       classifiers          = [
           'Development Status :: 3 - Alpha',
           'License :: OSI Approved :: MIT License',
           'Programming Language :: Python :: 2.7',
           'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)',
       ],
       keywords             = 'myhdl pihdf fpga',
       url                  = 'https://github.com/hnikolov/pihdf.git',
       author               = 'Hristo Nikolov, Nikolay Kavaldjiev',
       author_email         = 'h.n.nikolov@gmail.com, nikolay.kavaldjiev@gmail.com',
       license              = 'MIT',
       packages             = [
           'pihdf',
           'pihdf.interfaces',
           'pihdf.printers',
           'pihdf.bin'
       ],
       install_requires     = [
           'myhdl',
           'myhdl_lib',
           'simplejson',
           'coverage',
           'nose'
       ],
       test_suite           = 'nose.collector',
       tests_require        = ['nose', 'nose-cover3'],
       entry_points         = {
           'console_scripts': ['module=pihdf.bin.command_line:main']
       },
       include_package_data = True,
       zip_safe             = False
    )



from Common import Reset, Clock, Parameter
from STAvln import STAvln
from Bus import Bus
from HSD import HSD
from data_fields import net_fields, udp_fields, tcp_fields, ipv_fields, group_fields, port_fields, tcp_cmd_fields, payload_fields, test_fields, AXI4LM, AXI4LS, wd_fields, SBUS

__all__ = ["Reset",
           "Clock",
           "Parameter",
           "STAvln",
           "Bus",
           "HSD",
           "net_fields",
           "udp_fields",
           "tcp_fields",
           "ipv_fields",
           "group_fields",
           "port_fields",
           "tcp_cmd_fields",
           "payload_fields",
           "test_fields",
           "AXI4LM",
           "AXI4LS",
           "wd_fields",
           "SBUS"
          ]
